from . import sqlite3_lci_db
from .. import config, databases, geomapping
from ..data_store import ProcessedDataStore
from ..errors import UntypedExchange, InvalidExchange, UnknownObject, WrongDatabase
from ..project import writable_project
from ..query import Query
from ..search import IndexManager, Searcher
from ..utils import as_uncertainty_dict
from .proxies import Activity
from .schema import ActivityDataset, ExchangeDataset, get_id
from .utils import (
    check_exchange,
    get_csv_data_dict,
    dict_as_activitydataset,
    dict_as_exchangedataset,
    retupleize_geo_strings,
)
from bw_processing import clean_datapackage_name, create_datapackage
from fs.zipfs import ZipFS
from peewee import fn, DoesNotExist
import copy
import datetime
import itertools
import pandas
import pickle
import pprint
import pyprind
import random
import sqlite3
import warnings


_VALID_KEYS = {"location", "name", "product", "type"}


class SQLiteBackend(ProcessedDataStore):
    """
    A base class for SQLite backends.

    Subclasses must support at least the following calls:

    * ``load()``
    * ``write(data)``

    In addition, they should specify their backend with the ``backend`` attribute (a unicode string).

    * ``rename``
    * ``copy``
    * ``find_dependents``
    * ``random``
    * ``process``

    For new classes to be recognized by the ``DatabaseChooser``, they need to be registered with the ``config`` object, e.g.:

    .. code-block:: python

        config.backends['backend type string'] = BackendClass

    Instantiation does not load any data. If this database is not yet registered in the metadata store, a warning is written to ``stdout``.

    The data schema for databases in voluptuous is:

    .. code-block:: python

        exchange = {
                Required("input"): valid_tuple,
                Required("type"): basestring,
                }
        exchange.update(uncertainty_dict)
        lci_dataset = {
            Optional("categories"): Any(list, tuple),
            Optional("location"): object,
            Optional("unit"): basestring,
            Optional("name"): basestring,
            Optional("type"): basestring,
            Optional("exchanges"): [exchange]
        }
        db_validator = Schema({valid_tuple: lci_dataset}, extra=True)

    where:
        * ``valid_tuple`` is a :ref:`dataset identifier <dataset-codes>`, like ``("ecoinvent", "super strong steel")``
        * ``uncertainty_fields`` are fields from an :ref:`uncertainty dictionary <uncertainty-type>`.

    Processing a Database actually produces two parameter arrays: one for the exchanges, which make up the technosphere and biosphere matrices, and a geomapping array which links activities to locations.

    Args:
        *name* (unicode string): Name of the database to manage.

    """

    _metadata = databases
    validator = None
    backend = "sqlite"

    def __init__(self, *args, **kwargs):
        super(SQLiteBackend, self).__init__(*args, **kwargs)

        self._filters = {}
        self._order_by = None

    ### Generic LCI backend methods
    ###############################

    def copy(self, name):
        """Make a copy of the database.

        Internal links within the database will be updated to match the new database name, i.e. ``("old name", "some id")`` will be converted to ``("new name", "some id")`` for all exchanges.

        Args:
            * *name* (str): Name of the new database. Must not already exist.

        """
        assert name not in databases, ValueError("This database exists")
        data = self.relabel_data(copy.deepcopy(self.load()), name)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            new_database = self.__class__(name)
            new_database.register(format="Brightway2 copy",)

        new_database.write(data)
        return new_database

    def filepath_intermediate(self):
        raise NotImplementedError

    def filepath_processed(self):
        if self.metadata.get("dirty"):
            self.process()
        return self.dirpath_processed() / self.filename_processed()

    def find_dependents(self, data=None, ignore=None):
        """Get sorted list of direct dependent databases (databases linked from exchanges).

        Args:
            * *data* (dict, optional): Inventory data
            * *ignore* (list): List of database names to ignore

        Returns:
            List of database names

        """
        ignore = set(ignore if ignore is not None else [])
        if data is None:
            data = self.load()
            ignore.add(self.name)
        dependents = {
            exc.get("input")[0]
            for ds in data.values()
            for exc in ds.get("exchanges", [])
            if ds.get("type", "process") == "process"
            and exc.get("type") != "unknown"
            and exc.get("input", [None])[0] is not None
            and exc.get("input", [None])[0] not in ignore
        }
        return sorted(dependents)

    def find_graph_dependents(self):
        """Recursively get list of all dependent databases.

        Returns:
            A set of database names

        """

        def extend(seeds):
            return set.union(
                seeds, set.union(*[set(databases[obj]["depends"]) for obj in seeds])
            )

        seed, extended = {self.name}, extend({self.name})
        while extended != seed:
            seed, extended = extended, extend(extended)
        return extended

    def query(self, *queries):
        """Search through the database."""
        return Query(*queries)(self.load())

    def register(self, **kwargs):
        """Register a database with the metadata store.

        Databases must be registered before data can be written.

        Writing data automatically sets the following metadata:
            * *depends*: Names of the databases that this database references, e.g. "biosphere"
            * *number*: Number of processes in this database.

        Args:
            * *format* (str, optional): Format that the database was converted from, e.g. "Ecospold"

        """
        if "depends" not in kwargs:
            kwargs["depends"] = []
        kwargs["backend"] = self.backend
        super().register(**kwargs)

    def relabel_data(self, data, new_name):
        """Relabel database keys and exchanges.

        In a database which internally refer to the same database, update to new database name ``new_name``.

        Needed to copy a database completely or cut out a section of a database.

        For example:

        .. code-block:: python

            data = {
                ("old and boring", 1):
                    {"exchanges": [
                        {"input": ("old and boring", 42),
                        "amount": 1.0},
                        ]
                    },
                ("old and boring", 2):
                    {"exchanges": [
                        {"input": ("old and boring", 1),
                        "amount": 4.0}
                        ]
                    }
                }
            print(relabel_database(data, "shiny new"))
            >> {
                ("shiny new", 1):
                    {"exchanges": [
                        {"input": ("old and boring", 42),
                        "amount": 1.0},
                        ]
                    },
                ("shiny new", 2):
                    {"exchanges": [
                        {"input": ("shiny new", 1),
                        "amount": 4.0}
                        ]
                    }
                }

        In the example, the exchange to ``("old and boring", 42)`` does not change, as this is not part of the updated data.

        Args:
            * *data* (dict): The data to modify
            * *new_name* (str): The name of the modified database

        Returns:
            The modified data

        """

        def relabel_exchanges(obj, new_name):
            for e in obj.get("exchanges", []):
                if e["input"] in data:
                    e["input"] = (new_name, e["input"][1])
            return obj

        return dict(
            [
                ((new_name, k[1]), relabel_exchanges(v, new_name))
                for k, v in data.items()
            ]
        )

    def rename(self, name):
        """Rename a database. Modifies exchanges to link to new name. Deregisters old database.

        Args:
            * *name* (str): New name.

        Returns:
            New ``Database`` object.

        """
        old_name = self.name
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            new_db = self.__class__(name)
            databases[name] = databases[old_name]
        new_data = self.relabel_data(self.load(), name)
        new_db.write(new_data)
        del databases[old_name]
        self.name = name
        return new_db

    ### Iteration, filtering, and ordering
    ######################################

    # Private methods

    def __iter__(self):
        for ds in self._get_queryset():
            yield Activity(ds)

    def __len__(self):
        return self._get_queryset().count()

    def __contains__(self, obj):
        return self._get_queryset(filters={"code": obj[1]}).count() > 0

    @property
    def _searchable(self):
        return databases.get(self.name, {}).get("searchable", False)

    def _get_queryset(self, random=False, filters=True):
        qs = ActivityDataset.select().where(ActivityDataset.database == self.name)
        if filters:
            if isinstance(filters, dict):
                for key, value in filters.items():
                    qs = qs.where(getattr(ActivityDataset, key) == value)
            if self.filters:
                print("Using the following database filters:")
                pprint.pprint(self.filters)
                for key, value in self.filters.items():
                    qs = qs.where(getattr(ActivityDataset, key) == value)
        if self.order_by and not random:
            qs = qs.order_by(getattr(ActivityDataset, self.order_by))
        else:
            qs = qs.order_by(fn.Random())
        return qs

    def _get_filters(self):
        return self._filters

    def _set_filters(self, filters):
        if not filters:
            self._filters = {}
        else:
            print(
                "Filters will effect all database queries"
                " until unset (`.filters = None`)"
            )
            assert isinstance(filters, dict), "Filter must be a dictionary"
            for key in filters:
                assert key in _VALID_KEYS, "Filter key {} is invalid".format(key)
                self._filters = filters
        return self

    def _get_order_by(self):
        return self._order_by

    def _set_order_by(self, field):
        if not field:
            self._order_by = None
        else:
            assert field in _VALID_KEYS, "order_by field {} is invalid".format(field)
            self._order_by = field
        return self

    # Public API

    filters = property(_get_filters, _set_filters)
    order_by = property(_get_order_by, _set_order_by)

    def random(self, filters=True, true_random=False):
        """True random requires loading and sorting data in SQLite, and can be resource-intensive."""
        try:
            if true_random:
                return Activity(self._get_queryset(random=True, filters=filters).get())
            else:
                return Activity(
                    self._get_queryset(filters=filters)
                    .offset(random.randint(0, len(self)))
                    .get()
                )
        except DoesNotExist:
            warnings.warn("This database is empty")
            return None

    def get(self, code):
        if isinstance(code, int):
            return Activity(
                self._get_queryset(filters=False)
                .where(ActivityDataset.id == code)
                .get()
            )
        else:
            return Activity(
                self._get_queryset(filters=False)
                .where(ActivityDataset.code == code)
                .get()
            )

    ### Data management
    ###################

    # Private methods

    def _drop_indices(self):
        with sqlite3_lci_db.transaction():
            sqlite3_lci_db.execute_sql('DROP INDEX IF EXISTS "activitydataset_key"')
            sqlite3_lci_db.execute_sql('DROP INDEX IF EXISTS "exchangedataset_input"')
            sqlite3_lci_db.execute_sql('DROP INDEX IF EXISTS "exchangedataset_output"')

    def _add_indices(self):
        with sqlite3_lci_db.transaction():
            sqlite3_lci_db.execute_sql(
                'CREATE UNIQUE INDEX IF NOT EXISTS "activitydataset_key" ON "activitydataset" ("database", "code")'
            )
            sqlite3_lci_db.execute_sql(
                'CREATE INDEX IF NOT EXISTS "exchangedataset_input" ON "exchangedataset" ("input_database", "input_code")'
            )
            sqlite3_lci_db.execute_sql(
                'CREATE INDEX IF NOT EXISTS "exchangedataset_output" ON "exchangedataset" ("output_database", "output_code")'
            )

    def _efficient_write_dataset(self, index, key, ds, exchanges, activities):
        for exchange in ds.get("exchanges", []):
            if "input" not in exchange or "amount" not in exchange:
                raise InvalidExchange
            if "type" not in exchange:
                raise UntypedExchange
            exchange["output"] = key
            exchanges.append(dict_as_exchangedataset(exchange))

            # Query gets passed as INSERT INTO x VALUES ('?', '?'...)
            # SQLite3 has a limit of 999 variables,
            # So 6 fields * 125 is under the limit
            # Otherwise get the following:
            # peewee.OperationalError: too many SQL variables
            if len(exchanges) > 125:
                ExchangeDataset.insert_many(exchanges).execute()
                exchanges = []

        ds = {k: v for k, v in ds.items() if k != "exchanges"}
        ds["database"] = key[0]
        ds["code"] = key[1]

        activities.append(dict_as_activitydataset(ds))

        if len(activities) > 125:
            ActivityDataset.insert_many(activities).execute()
            activities = []

        if not getattr(config, "is_test", None):
            self.pbar.update()

        return exchanges, activities

    def _efficient_write_many_data(self, data, indices=True):
        be_complicated = len(data) >= 100 and indices
        if be_complicated:
            self._drop_indices()
        sqlite3_lci_db.db.autocommit = False
        try:
            sqlite3_lci_db.db.begin()
            self.delete(keep_params=True, warn=False)
            exchanges, activities = [], []

            if not getattr(config, "is_test", None):
                self.pbar = pyprind.ProgBar(
                    len(data),
                    title="Writing activities to SQLite3 database:",
                    monitor=True,
                )

            for index, (key, ds) in enumerate(data.items()):
                exchanges, activities = self._efficient_write_dataset(
                    index, key, ds, exchanges, activities
                )

            if not getattr(config, "is_test", None):
                print(self.pbar)
                del self.pbar

            if activities:
                ActivityDataset.insert_many(activities).execute()
            if exchanges:
                ExchangeDataset.insert_many(exchanges).execute()
            sqlite3_lci_db.db.commit()
        except:
            sqlite3_lci_db.db.rollback()
            raise
        finally:
            sqlite3_lci_db.db.autocommit = True
            if be_complicated:
                self._add_indices()

    # Public API

    @writable_project
    def write(self, data, process=True):
        """Write ``data`` to database.

        ``data`` must be a dictionary of the form::

            {
                ('database name', 'dataset code'): {dataset}
            }

        Writing a database will first deletes all existing data."""
        if self.name not in databases:
            self.register()
        wrong_database = {key[0] for key in data}.difference({self.name})
        if wrong_database:
            raise WrongDatabase(
                "Can't write activities in databases {} to database {}".format(
                    wrong_database, self.name
                )
            )

        databases[self.name]["number"] = len(data)
        databases.set_modified(self.name)

        geomapping.add({x["location"] for x in data.values() if x.get("location")})
        if data:
            try:
                self._efficient_write_many_data(data)
            except:
                # Purge all data from database, then reraise
                self.delete(warn=False)
                raise

        self.make_searchable(reset=True)

        if process:
            self.process()

    def load(self, *args, **kwargs):
        # Should not be used, in general; relatively slow
        activities = [obj["data"] for obj in self._get_queryset().dicts()]

        activities = {(o["database"], o["code"]): o for o in activities}
        for o in activities.values():
            o["exchanges"] = []

        exchange_qs = (
            ExchangeDataset.select(ExchangeDataset.data)
            .where(ExchangeDataset.output_database == self.name)
            .dicts()
        )

        for exc in exchange_qs:
            try:
                activities[exc["data"]["output"]]["exchanges"].append(exc["data"])
            except KeyError:
                # This exchange not in the reduced set of activities returned
                # by _get_queryset
                pass
        return activities

    def new_activity(self, code, **kwargs):
        obj = Activity()
        obj["database"] = self.name
        obj["code"] = str(code)
        obj["location"] = config.global_location
        obj.update(kwargs)
        return obj

    @writable_project
    def make_searchable(self, reset=False):
        if self.name not in databases:
            raise UnknownObject("This database is not yet registered")
        if self._searchable and not reset:
            print("This database is already searchable")
            return
        databases[self.name]["searchable"] = True
        databases.flush()
        IndexManager(self.filename).delete_database()
        IndexManager(self.filename).add_datasets(self)

    @writable_project
    def make_unsearchable(self):
        databases[self.name]["searchable"] = False
        databases.flush()
        IndexManager(self.filename).delete_database()

    @writable_project
    def delete(self, keep_params=False, warn=True):
        """Delete all data from SQLite database and Whoosh index"""
        if warn:
            MESSAGE = """
            Please use `del databases['{}']` instead.
            Otherwise, the metadata and database get out of sync.
            Call `.delete(warn=False)` to skip this message in the future.
            """
            warnings.warn(MESSAGE.format(self.name), UserWarning)

        vacuum_needed = len(self) > 500

        ActivityDataset.delete().where(ActivityDataset.database == self.name).execute()
        ExchangeDataset.delete().where(
            ExchangeDataset.output_database == self.name
        ).execute()
        IndexManager(self.filename).delete_database()

        if not keep_params:
            from ..parameters import (
                DatabaseParameter,
                ActivityParameter,
                ParameterizedExchange,
            )

            groups = tuple(
                {
                    o[0]
                    for o in ActivityParameter.select(ActivityParameter.group)
                    .where(ActivityParameter.database == self.name)
                    .tuples()
                }
            )
            ParameterizedExchange.delete().where(
                ParameterizedExchange.group << groups
            ).execute()
            ActivityParameter.delete().where(
                ActivityParameter.database == self.name
            ).execute()
            DatabaseParameter.delete().where(
                DatabaseParameter.database == self.name
            ).execute()

        if vacuum_needed:
            sqlite3_lci_db.vacuum()

    def exchange_data_iterator(self, sql, dependents, flip=False):
        """Iterate over exchanges and format for ``bw_processing`` arrays.

        ``dependents`` is a set of dependent database names.

        ``flip`` means flip the numeric sign; see ``bw_processing`` docs.

        Uses raw sqlite3 to retrieve data for ~2x speed boost."""
        connection = sqlite3.connect(sqlite3_lci_db._filepath)
        cursor = connection.cursor()
        for row in cursor.execute(sql, (self.name,)):
            data, row, col, input_database, input_code, output_database, output_code = row
            # Modify ``dependents`` in place
            if input_database != output_database:
                dependents.add(input_database)
            data = pickle.loads(bytes(data))
            check_exchange(data)
            if row is None or col is None:
                raise UnknownObject(
                    (
                        "Exchange between {} and {} is invalid "
                        "- one of these objects is unknown (i.e. doesn't exist "
                        "as a process dataset)"
                    ).format(
                        (input_database, input_code), (output_database, output_code)
                    )
                )
            yield {
                **as_uncertainty_dict(data),
                "row": row,
                "col": col,
                "flip": flip,
            }

    def process(self, csv=False):
        """Create structured arrays for the technosphere and biosphere matrices.

        Uses ``bw_processing`` for array creation and metadata serialization.

        Also creates a ``geomapping`` array, linking activities to locations. Used for regionalized calculations.

        Use a raw SQLite3 cursor instead of Peewee for a ~2 times speed advantage.

        """
        # Try to avoid race conditions - but no guarantee
        self.metadata["processed"] = datetime.datetime.now().isoformat()

        # Get number of exchanges and processes to set
        # initial Numpy array size (still have to include)
        # implicit production exchanges
        dependents = set()

        # Create geomapping array, from dataset interger ids to locations
        inv_mapping_qs = ActivityDataset.select(
            ActivityDataset.id, ActivityDataset.location
        ).where(
            ActivityDataset.database == self.name, ActivityDataset.type == "process"
        )

        # self.filepath_processed checks if data is dirty,
        # and processes if it is. This causes an infinite loop.
        # So we construct the filepath ourselves.
        fp = str(self.dirpath_processed() / self.filename_processed())

        dp = create_datapackage(
            fs=ZipFS(fp, write=True),
            name=clean_datapackage_name(self.name),
            sum_intra_duplicates=True,
            sum_inter_duplicates=False,
        )
        dp.add_persistent_vector_from_iterator(
            matrix="inv_geomapping_matrix",
            name=clean_datapackage_name(self.name + " inventory geomapping matrix"),
            dict_iterator=(
                {
                    "row": row[0],
                    "col": geomapping[
                        retupleize_geo_strings(row[1])
                        or config.global_location
                    ],
                    "amount": 1,
                }
                for row in inv_mapping_qs.tuples()
            ),
            nrows=inv_mapping_qs.count(),
        )

        BIOSPHERE_SQL = """SELECT e.data, a.id, b.id, e.input_database, e.input_code, e.output_database, e.output_code
                FROM exchangedataset as e
                LEFT JOIN activitydataset as a ON a.code == e.input_code AND a.database == e.input_database
                LEFT JOIN activitydataset as b ON b.code == e.output_code AND b.database == e.output_database
                WHERE e.output_database = ?
                AND e.type = 'biosphere'
        """
        dp.add_persistent_vector_from_iterator(
            matrix="biosphere_matrix",
            name=clean_datapackage_name(self.name + " biosphere matrix"),
            dict_iterator=self.exchange_data_iterator(BIOSPHERE_SQL, dependents),
        )

        # Figure out when the production exchanges are implicit
        implicit_production = (
            {"row": get_id((self.name, x[0])), "amount": 1}
            # Get all codes
            for x in ActivityDataset.select(ActivityDataset.code)
            .where(
                # Get correct database name
                ActivityDataset.database == self.name,
                # Only consider `process` type activities
                ActivityDataset.type << ("process", None),
                # But exclude activities that already have production exchanges
                ~(
                    ActivityDataset.code
                    << ExchangeDataset.select(
                        # Get codes to exclude
                        ExchangeDataset.output_code
                    ).where(
                        ExchangeDataset.output_database == self.name,
                        ExchangeDataset.type << ("production", "generic production"),
                    )
                ),
            )
            .tuples()
        )

        TECHNOSPHERE_POSITIVE_SQL = """SELECT e.data, a.id, b.id, e.input_database, e.input_code, e.output_database, e.output_code
                FROM exchangedataset as e
                LEFT JOIN activitydataset as a ON a.code == e.input_code AND a.database == e.input_database
                LEFT JOIN activitydataset as b ON b.code == e.output_code AND b.database == e.output_database
                WHERE e.output_database = ?
                AND e.type IN ('production', 'substitution', 'generic production')
        """
        TECHNOSPHERE_NEGATIVE_SQL = """SELECT e.data, a.id, b.id, e.input_database, e.input_code, e.output_database, e.output_code
                FROM exchangedataset as e
                LEFT JOIN activitydataset as a ON a.code == e.input_code AND a.database == e.input_database
                LEFT JOIN activitydataset as b ON b.code == e.output_code AND b.database == e.output_database
                WHERE e.output_database = ?
                AND e.type IN ('technosphere', 'generic consumption')"""

        dp.add_persistent_vector_from_iterator(
            matrix="technosphere_matrix",
            name=clean_datapackage_name(self.name + " technosphere matrix"),
            dict_iterator=itertools.chain(
                self.exchange_data_iterator(
                    TECHNOSPHERE_NEGATIVE_SQL, dependents, flip=True
                ),
                self.exchange_data_iterator(TECHNOSPHERE_POSITIVE_SQL, dependents),
                implicit_production,
            ),
        )
        if csv:
            df = pandas.DataFrame([get_csv_data_dict(ds) for ds in self])
            dp.add_csv_metadata(
                dataframe=df,
                valid_for=[
                    (
                        clean_datapackage_name(self.name + " technosphere matrix"),
                        "cols",
                    ),
                    (clean_datapackage_name(self.name + " biosphere matrix"), "cols"),
                ],
                name=clean_datapackage_name(self.name + " activity metadata"),
            )

        dp.finalize_serialization()

        self.metadata["depends"] = sorted(dependents)
        self.metadata["dirty"] = False
        self._metadata.flush()

    def search(self, string, **kwargs):
        """Search this database for ``string``.

        The searcher include the following fields:

        * name
        * comment
        * categories
        * location
        * reference product

        ``string`` can include wild cards, e.g. ``"trans*"``.

        By default, the ``name`` field is given the most weight. The full weighting set is called the ``boost`` dictionary, and the default weights are::

            {
                "name": 5,
                "comment": 1,
                "product": 3,
                "categories": 2,
                "location": 3
            }

        Optional keyword arguments:

        * ``limit``: Number of results to return.
        * ``boosts``: Dictionary of field names and numeric boosts - see default boost values above. New values must be in the same format, but with different weights.
        * ``filter``: Dictionary of criteria that search results must meet, e.g. ``{'categories': 'air'}``. Keys must be one of the above fields.
        * ``mask``: Dictionary of criteria that exclude search results. Same format as ``filter``.
        * ``facet``: Field to facet results. Must be one of ``name``, ``product``, ``categories``, ``location``, or ``database``.
        * ``proxy``: Return ``Activity`` proxies instead of raw Whoosh documents. Default is ``True``.

        Returns a list of ``Activity`` datasets."""
        with Searcher(self.filename) as s:
            results = s.search(string, **kwargs)
        return results

    def graph_technosphere(self, filename=None, **kwargs):
        from bw2analyzer.matrix_grapher import SparseMatrixGrapher
        from bw2calc import LCA

        lca = LCA({self.random(): 1})
        lca.lci()

        smg = SparseMatrixGrapher(lca.technosphere_matrix)
        return smg.ordered_graph(filename, **kwargs)



from . import config
from .errors import ReadOnlyProject
from .filesystem import create_dir
from .sqlite import PickleField, SubstitutableDatabase
from .utils import maybe_path
from bw_processing import safe_filename
from fasteners import InterProcessLock
from collections.abc import Iterable
from pathlib import Path
from peewee import Model, TextField, BooleanField, DoesNotExist
from threading import ThreadError
import appdirs
import os
import shutil
import tempfile
import warnings
import wrapt


READ_ONLY_PROJECT = """
***Read only project***

This project is being used by another process and no writes can be made until:
    1. You close the other program, or switch to a different project, *and*
    2. You call `projects.enable_writes` *and* get the response `True`.

    If you are **sure** that this warning is incorrect, call
    `projects.enable_writes(force=True)` to enable writes.
"""


def lockable():
    return hasattr(config, "p") and config.p.get("lockable")


class ProjectDataset(Model):
    data = PickleField()
    name = TextField(index=True, unique=True)
    full_hash = BooleanField(default=True)

    def __str__(self):
        return "Project: {}".format(self.name)

    __repr__ = lambda x: str(x)

    def __lt__(self, other):
        if not isinstance(other, ProjectDataset):
            raise TypeError
        else:
            return self.name.lower() < other.name.lower()


class ProjectManager(Iterable):
    _basic_directories = (
        "backups",
        "intermediate",
        "lci",
        "processed",
    )
    _is_temp_dir = False
    read_only = False

    def __init__(self):
        self._base_data_dir, self._base_logs_dir = self._get_base_directories()
        self._create_base_directories()
        self.db = SubstitutableDatabase(
            self._base_data_dir / "projects.db", [ProjectDataset]
        )

        columns = {o.name for o in self.db._database.get_columns("projectdataset")}
        if "full_hash" not in columns:
            src_filepath = self._base_data_dir / "projects.db"
            backup_filepath = self._base_data_dir / "projects.backup.db"
            shutil.copy(src_filepath, backup_filepath)

            MIGRATION_WARNING = """Adding a column to the projects database. A backup copy of this database '{}' was made at '{}'; if you have problems, file an issue, and restore the backup data to use the stable version of Brightway2."""

            print(MIGRATION_WARNING.format(src_filepath, backup_filepath))

            ADD_FULL_HASH_COLUMN = """ALTER TABLE projectdataset ADD COLUMN "full_hash" integer default 1"""
            self.db.execute_sql(ADD_FULL_HASH_COLUMN)

            # We don't do this, as the column added doesn't have a default
            # value, meaning that one would get error from using the
            # development branch alongside the stable branch.

            # from playhouse.migrate import SqliteMigrator, migrate
            # migrator = SqliteMigrator(self.db._database)
            # full_hash = BooleanField(default=True)
            # migrate(migrator.add_column("projectdataset", "full_hash", full_hash),)
        self.set_current("default", update=False)

    def __iter__(self):
        for project_ds in ProjectDataset.select():
            yield project_ds

    def __contains__(self, name):
        return ProjectDataset.select().where(ProjectDataset.name == name).count() > 0

    def __len__(self):
        return ProjectDataset.select().count()

    def __repr__(self):
        if len(self) > 20:
            return (
                "Brightway2 projects manager with {} objects, including:"
                "{}\nUse `sorted(projects)` to get full list, "
                "`projects.report()` to get\n\ta report on all projects."
            ).format(
                len(self),
                "".join(
                    ["\n\t{}".format(x) for x in sorted([x.name for x in self])[:10]]
                ),
            )
        else:
            return (
                "Brightway2 projects manager with {} objects:{}"
                "\nUse `projects.report()` to get a report on all projects."
            ).format(
                len(self),
                "".join(["\n\t{}".format(x) for x in sorted([x.name for x in self])]),
            )

    ### Internal functions for managing projects

    def _get_base_directories(self):
        envvar = maybe_path(os.getenv("BRIGHTWAY2_DIR"))
        if envvar:
            if not envvar.is_dir():
                raise OSError(
                    (
                        "BRIGHTWAY2_DIR variable is {}, but this is not"
                        " a valid directory"
                    ).format(envvar)
                )
            else:
                print(
                    "Using environment variable BRIGHTWAY2_DIR for data "
                    "directory:\n{}".format(envvar)
                )
                envvar = envvar.absolute()
                logs_dir = envvar / "logs"
                create_dir(logs_dir)
                return envvar, logs_dir

        LABEL = "Brightway3"
        data_dir = Path(appdirs.user_data_dir(LABEL, "pylca"))
        logs_dir = Path(appdirs.user_log_dir(LABEL, "pylca"))
        return data_dir, logs_dir

    def _create_base_directories(self):
        create_dir(self._base_data_dir)
        create_dir(self._base_logs_dir)

    @property
    def current(self):
        return self._project_name

    @property
    def twofive(self):
        return bool(self.dataset.data.get("25"))

    def set_current(self, name, writable=True, update=True):
        if not self.read_only and lockable() and hasattr(self, "_lock"):
            try:
                self._lock.release()
            except (RuntimeError, ThreadError):
                pass
        self._project_name = str(name)

        # Need to allow writes when creating a new project
        # for new metadata stores
        self.read_only = False
        self.create_project(name)
        self.dataset = ProjectDataset.get(ProjectDataset.name == self._project_name)
        self._reset_meta()
        self._reset_sqlite3_databases()

        self.dataset = ProjectDataset.get(name=name)

        if not lockable():
            pass
        elif writable:
            self._lock = InterProcessLock(self.dir / "write-lock")
            self.read_only = not self._lock.acquire(timeout=0.05)
            if self.read_only:
                warnings.warn(READ_ONLY_PROJECT)
        else:
            self.read_only = True

        if not self.read_only and update:
            self._do_automatic_updates()

    def _do_automatic_updates(self):
        """Run any available automatic updates"""
        from .updates import Updates

        for update_name in Updates.check_automatic_updates():
            print("Applying automatic update: {}".format(update_name))
            Updates.do_update(update_name)

    def _reset_meta(self):
        for obj in config.metadata:
            obj.__init__()

    def _reset_sqlite3_databases(self):
        for relative_path, substitutable_db in config.sqlite3_databases:
            substitutable_db.change_path(self.dir / relative_path)

    ### Public API
    @property
    def dir(self):
        return Path(self._base_data_dir) / safe_filename(
            self.current, full=self.dataset.full_hash
        )

    @property
    def logs_dir(self):
        return Path(self._base_logs_dir) / safe_filename(
            self.current, full=self.dataset.full_hash
        )

    @property
    def output_dir(self):
        """Get directory for output files.

        Uses environment variable ``BRIGHTWAY2_OUTPUT_DIR``; ``preferences['output_dir']``; or directory ``output`` in current project.

        Returns output directory path.

        """
        ep, pp = (
            maybe_path(os.getenv("BRIGHTWAY2_OUTPUT_DIR")),
            maybe_path(config.p.get("output_dir")),
        )
        if ep and ep.is_dir():
            return ep
        elif pp and pp.is_dir():
            return pp
        else:
            return self.request_directory("output")

    def create_project(self, name=None, **kwargs):
        name = name or self.current

        kwargs["25"] = True
        full_hash = kwargs.pop("full_hash", False)
        try:
            self.dataset = ProjectDataset.get(ProjectDataset.name == name)
        except DoesNotExist:
            self.dataset = ProjectDataset.create(
                data=kwargs, name=name, full_hash=full_hash
            )
        create_dir(self.dir)
        for dir_name in self._basic_directories:
            create_dir(self.dir / dir_name)
        create_dir(self.logs_dir)

    def copy_project(self, new_name, switch=True):
        """Copy current project to a new project named ``new_name``. If ``switch``, switch to new project."""
        if new_name in self:
            raise ValueError("Project {} already exists".format(new_name))
        fp = self._base_data_dir / safe_filename(new_name, full=self.dataset.full_hash)
        if fp.exists():
            raise ValueError("Project directory already exists")
        project_data = ProjectDataset.get(ProjectDataset.name == self.current).data
        ProjectDataset.create(
            data=project_data, name=new_name, full_hash=self.dataset.full_hash
        )
        shutil.copytree(self.dir, fp, ignore=lambda x, y: ["write-lock"])
        create_dir(self._base_logs_dir / safe_filename(new_name))
        if switch:
            self.set_current(new_name)

    def request_directory(self, name):
        """Return the absolute path to the subdirectory ``dirname``, creating it if necessary.

        Returns ``False`` if directory can't be created."""
        fp = self.dir / str(name)
        create_dir(fp)
        if not fp.is_dir():
            return False
        return fp

    def _use_temp_directory(self):
        """Point the ProjectManager towards a temporary directory instead of `user_data_dir`.

        Used exclusively for tests."""
        if not self._is_temp_dir:
            self._orig_base_data_dir = self._base_data_dir
            self._orig_base_logs_dir = self._base_logs_dir
        temp_dir = Path(tempfile.mkdtemp())
        self._base_data_dir = temp_dir / "data"
        self._base_logs_dir = temp_dir / "logs"
        self.db.change_path(":memory:")
        self.set_current("default", update=False)
        self._is_temp_dir = True
        return temp_dir

    def _restore_orig_directory(self):
        """Point the ProjectManager back to original directories.

        Used exclusively in tests."""
        if not self._is_temp_dir:
            return
        self._base_data_dir = self._orig_base_data_dir
        del self._orig_base_data_dir
        self._base_logs_dir = self._orig_base_logs_dir
        del self._orig_base_logs_dir
        self.db.change_path(self._base_data_dir / "projects.db")
        self.set_current("default", update=False)
        self._is_temp_dir = False

    def migrate_project_25(self):
        """Migrate project to Brightway 2.5.

        Reprocesses all databases and LCIA objects."""
        assert not self.twofive, "Project is already 2.5 compatible"

        from .updates import Updates
        Updates()._reprocess_all()

        self.dataset.data["25"] = True
        self.dataset.save()

    def delete_project(self, name=None, delete_dir=False):
        """Delete project ``name``, or the current project.

        ``name`` is the project to delete. If ``name`` is not provided, delete the current project.

        By default, the underlying project directory is not deleted; only the project name is removed from the list of active projects. If ``delete_dir`` is ``True``, then also delete the project directory.

        If deleting the current project, this function sets the current directory to ``default`` if it exists, or to a random project.

        Returns the current project."""
        victim = name or self.current
        if victim not in self:
            raise ValueError("{} is not a project".format(victim))

        if len(self) == 1:
            raise ValueError("Can't delete only remaining project")

        ProjectDataset.delete().where(ProjectDataset.name == victim).execute()

        if delete_dir:
            dir_path = self._base_data_dir / safe_filename(victim)
            assert dir_path.is_dir(), "Can't find project directory"
            shutil.rmtree(dir_path)

        if name is None or name == self.current:
            if "default" in self:
                self.set_current("default")
            else:
                self.set_current(next(iter(self)).name)
        return self.current

    def purge_deleted_directories(self):
        """Delete project directories for projects which are no longer registered.

        Returns number of directories deleted."""
        registered = {safe_filename(obj.name) for obj in self}
        bad_directories = [
            self._base_data_dir / dirname
            for dirname in os.listdir(self._base_data_dir)
            if (self._base_data_dir / dirname).is_dir() and dirname not in registered
        ]

        for fp in bad_directories:
            shutil.rmtree(fp)

        return len(bad_directories)

    def report(self):
        """Give a report on current projects, including installed databases and file sizes.

        Returns tuples of ``(project name, number of databases, size of all databases (GB))``."""
        from . import databases

        _current = self.current
        data = []

        def get_dir_size(dirpath):
            """Modified from http://stackoverflow.com/questions/12480367/how-to-generate-directory-size-recursively-in-python-like-du-does.

            Does not follow symbolic links"""
            return sum(
                sum(os.path.getsize(root / name) for name in files)
                for root, dirs, files in os.walk(dirpath)
            )

        names = sorted([x.name for x in self])
        for obj in names:
            self.set_current(obj, update=False, writable=False)
            data.append((obj, len(databases), get_dir_size(projects.dir) / 1e9))
        self.set_current(_current)
        return data

    def use_short_hash(self):
        if not self.dataset.full_hash:
            return
        try:
            old_dir, old_logs_dir = self.dir, self.logs_dir
            self.dataset.full_hash = False
            if self.dir.exists():
                raise OSError("Target directory {} already exists".format(self.dir))
            if self.logs_dir.exists():
                raise OSError(
                    "Target directory {} already exists".format(self.logs_dir)
                )
            old_dir.rename(self.dir)
            old_logs_dir.rename(self.logs_dir)
            self.dataset.save()
        except Exception as ex:
            self.dataset.full_hash = True
            raise ex

    def use_full_hash(self):
        if self.dataset.full_hash:
            return
        try:
            old_dir, old_logs_dir = self.dir, self.logs_dir
            self.dataset.full_hash = True
            if self.dir.exists():
                raise OSError("Target directory {} already exists".format(self.dir))
            if self.logs_dir.exists():
                raise OSError(
                    "Target directory {} already exists".format(self.logs_dir)
                )
            old_dir.rename(self.dir)
            old_logs_dir.rename(self.logs_dir)
            self.dataset.save()
        except Exception as ex:
            self.dataset.full_hash = False
            raise ex


projects = ProjectManager()


@wrapt.decorator
def writable_project(wrapped, instance, args, kwargs):
    if projects.read_only:
        raise ReadOnlyProject(READ_ONLY_PROJECT)
    return wrapped(*args, **kwargs)



from bw2data.tests import bw2test
from bw2data import projects
from bw2data.data_store import DataStore, ProcessedDataStore
from bw2data.errors import UnknownObject
from bw2data.serialization import SerializedDict

# from bw_processing import load_package, COMMON_DTYPE
from numbers import Number
from voluptuous import Schema
import numpy as np
import os
import pickle
import pytest


class Metadata(SerializedDict):
    filename = "mock-meta.json"


metadata = Metadata()


class MockDS(DataStore):
    """Mock DataStore for testing"""

    _metadata = metadata
    validator = Schema(Number)


class MockPDS(ProcessedDataStore):
    """Mock DataStore for testing"""

    _metadata = metadata
    validator = Schema(Number)

    def process_row(self, row):
        return row


@pytest.fixture
@bw2test
def reset():
    metadata.__init__()


### DataStore


def test_data_store_repr(reset):
    d = MockDS("food")
    assert isinstance(str(d), str)


def test_data_store_unicode(reset):
    d = MockDS("food")
    assert str(d) == "Brightway2 MockDS: food"


def test_data_store_deregister(reset):
    d = MockDS("evening")
    d.register()
    assert "evening" in metadata
    d.deregister()
    assert "evening" not in metadata


def test_data_store_metadata_keyerror(reset):
    d = MockDS("evening")
    with pytest.raises(UnknownObject):
        d.metadata


def test_data_store_metadata_readable_writable(reset):
    d = MockDS("twilight")
    d.register()
    d.metadata = {"foo": "bar"}
    assert d.metadata == {"foo": "bar"}


def test_data_store_write_load(reset):
    d = MockDS("full moon")
    d.write(range(10))
    data = pickle.load(
        open(os.path.join(projects.dir, "intermediate", d.filename + ".pickle"), "rb")
    )
    assert list(data) == list(range(10))


def test_data_store_copy(reset):
    d = MockDS("full moon")
    d.register(foo="bar")
    d.write(range(10))
    gibbous = d.copy("waning gibbous")
    assert list(gibbous.load()) == list(range(10))
    assert metadata["waning gibbous"] == {"foo": "bar"}


def test_data_store_validation(reset):
    d = MockDS("cat")
    assert d.validate(4)



import collections
import operator
import itertools

operators = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "is": operator.eq,  # Note: not pure python `is`!
    "iis": lambda x, y: x.lower() == y.lower(),
    "!=": operator.ne,
    "<>": operator.ne,
    "not": operator.ne,
    "inot": lambda x, y: x.lower() != y.lower(),
    ">=": operator.ge,
    ">": operator.gt,
    "has": operator.contains,
    "ihas": lambda x, y: y.lower() in x.lower(),
    "nothas": lambda x, y: not operator.contains(x, y),
    "in": lambda x, y: operator.contains(y, x),
    "notin": lambda x, y: not operator.contains(y, x),
    # No iin because in normally doesn't take string inputs
    "len": lambda x, y: len(x) == y,
}


def try_op(f, x, y):
    try:
        return f(x, y)
    except:
        return False


class Dictionaries(object):
    """Pretends to be a single dictionary when applying a ``Query`` to multiple databases.

    Usage:
        first_database = Database(...).load()
        second_database = Database(...).load()
        my_joined_dataset = Dictionaries(first_database, second_database)
        search_results = Query(filter_1, filter_2)(my_joined_dataset)

    """

    def __init__(self, *args):
        self.dicts = args

    def items(self):
        return itertools.chain(*[x.items() for x in self.dicts])


class Result(object):
    """A container that wraps a filtered dataset. Returned by a calling a ``Query`` object. A result object functions like a read-only dictionary; you can call ``Result[some_key]``, or ``some_key in Result``, or ``len(Result)``.

    The dataset can also be sorted, using ``sort(field)``; the underlying data is then a ``collections.OrderedDict``.

    Args:
        * *result* (dict): The filtered dataset.

    """

    def __init__(self, result):
        self.result = result
        if not isinstance(result, dict):
            raise ValueError("Must pass dictionary")

    def __str__(self):
        return "Query result with %i entries" % len(self.result)

    def __repr__(self):
        if not self.result:
            return "Query result:\n\tNo query results found."
        data = list(self.result.items())[:20]
        return "Query result: (total %i)\n" % len(self.result) + "\n".join(
            ["%s: %s" % (k, v.get("name", "Unknown")) for k, v in data]
        )

    def sort(self, field, reverse=False):
        """Sort the filtered dataset. Operates in place; does not return anything.

        Args:
            * *field* (str): The key used for sorting.
            * *reverse* (bool, optional): Reverse normal sorting order.

        """
        self.result = collections.OrderedDict(
            sorted(
                self.result.items(),
                key=lambda t: t[1].get(field, None),
                reverse=reverse,
            )
        )

    # Generic dictionary methods
    def __len__(self):
        return len(self.result)

    def __iter__(self):
        return iter(self.result)

    def keys(self):
        return self.result.keys()

    def items(self):
        return self.result.items()

    def items(self):
        return self.result.items()

    def __getitem__(self, key):
        return self.result[key]

    def __contains__(self, key):
        return key in self.result


class Query(object):
    """A container for a set of filters applied to a dataset.

    Filters are applied by calling the ``Query`` object, and passing the dataset to filter as the argument. Calling a ``Query`` with some data returns a ``Result`` object with the filtered dataset.

    Args:
        * *filters* (filters): One or more ``Filter`` objects.

    """

    def __init__(self, *filters):
        self.filters = list(filters)

    def add(self, filter_):
        """Add another filter.

        Args:
            *filter_* (``Filter``): A Filter object.

        """
        self.filters.append(filter_)

    def __call__(self, data):
        for filter_ in self.filters:
            data = filter_(data)
        return Result(data)


class Filter(object):
    """A filter on a dataset.

    The following functions are supported:

        * "<", "<=", "==", ">", ">=": Mathematical relations
        * "is", "not": Identity relations. Work on any Python object.
        * "in", "notin": List or string relations.
        * "iin", "iis", "inot": Case-insensitive string relations.
        * "len": Length relation.

    In addition, any function which defines a relationship between an input and an output can also be used.

    Examples:

        * All ``name`` values are *"foo"*: ``Filter("name", "is", "foo")``
        * All ``name`` values include the string *"foo"*: ``Filter("name", "has", "foo")``
        * Category (a list of categories and subcategories) includes *"foo"*: ``Filter("category", "has", "foo")``

    Args:
        * *key* (str): The field to filter on.
        * *function* (str or object): One of the pre-defined filters, or a callable object.
        * *value* (object): The value to test against.

    Returns:
        A ``Result`` object which wraps a new data dictionary.

    """

    def __init__(self, key, function, value):
        self.key = key
        self.function = function
        self.value = value
        if not callable(function):
            self.function = operators.get(function, None)
        if not self.function:
            raise ValueError("No valid function found")

    def __call__(self, data):
        return dict(
            (
                (k, v)
                for k, v in data.items()
                if try_op(self.function, v.get(self.key, None), self.value)
            )
        )


def NF(value):
    """Shortcut for a name filter"""
    return Filter("name", "has", value)


def PF(value):
    """Shortcut for a reference product filter"""
    return Filter("reference product", "has", value)



from . import (
    Database,
    databases,
    Method,
    methods,
    Weighting,
    weightings,
    Normalization,
    normalizations,
    preferences,
    projects,
)
from .backends import sqlite3_lci_db
import numpy as np
import os
import pyprind
import re
import shutil
import sqlite3
import warnings
import pickle

hash_re = re.compile("^[a-zA-Z0-9]{32}$")
is_hash = lambda x: bool(hash_re.match(x))

UPTODATE_WARNING = "\n\nYour data needs to be updated. Please run the following program on the command line:\n\n\tbw2-uptodate\n"

UPDATE_ACTIVITYDATASET = """
BEGIN;
DROP INDEX IF EXISTS "activitydataset_key";
ALTER TABLE ActivityDataset rename to AD_old;
CREATE TABLE "activitydataset" (
    "id" INTEGER NOT NULL PRIMARY KEY,
    "database" TEXT NOT NULL,
    "code" TEXT NOT NULL,
    "data" BLOB NOT NULL,
    "location" TEXT,
    "name" TEXT,
    "product" TEXT,
    "type" TEXT
);
INSERT INTO ActivityDataset ("database", "code", "data", "location", "name", "product", "type")
    SELECT substr(key, 0, instr(key, '⊡')),
        substr("key", instr("key", '⊡') + 1),
        "data",
        "location",
        "name",
        "product",
        "type"
    FROM AD_old;
CREATE UNIQUE INDEX "activitydataset_key" ON "activitydataset" ("database", "code");
DROP TABLE AD_old;
COMMIT;
"""

UPDATE_EXCHANGEDATASET = """
BEGIN;
DROP INDEX IF EXISTS "exchangedataset_database";
DROP INDEX IF EXISTS "exchangedataset_input";
DROP INDEX IF EXISTS "exchangedataset_output";
ALTER TABLE ExchangeDataset rename to ED_old;
CREATE TABLE "exchangedataset" (
    "id" INTEGER NOT NULL PRIMARY KEY,
    "data" BLOB NOT NULL,
    "input_database" TEXT NOT NULL,
    "input_code" TEXT NOT NULL,
    "output_database" TEXT NOT NULL,
    "output_code" TEXT NOT NULL,
    "type" TEXT NOT NULL
);
INSERT INTO ExchangeDataset ("data", "input_database", "input_code", "output_database", "output_code", "type")
    SELECT "data",
        substr("input", 0, instr("input", '⊡')),
        substr("input", instr("input", '⊡') + 1),
        substr("output", 0, instr("output", '⊡')),
        substr("output", instr("output", '⊡') + 1),
        "type"
    FROM ED_old;
CREATE INDEX "exchangedataset_input" ON "exchangedataset" ("input_database", "input_code");
CREATE INDEX "exchangedataset_output" ON "exchangedataset" ("output_database", "output_code");
DROP TABLE ED_old;
COMMIT;
"""


class Updates(object):
    UPDATES = {
        "2.0 schema change": {
            "method": "schema_change_20_compound_keys",
            "explanation": "",
            "automatic": True,
        },
        "2.0-2 database search directories": {
            "method": "database_search_directories_20",
            "automatic": True,
            "explanation": "",
        },
        "2.3 processed data format": {
            "method": "processed_data_format_change_23",
            "automatic": True,
            "explanation": "",
        },
        # "4.0 new processed format": {
        #     'method': 'expire_all_processed_data_40',
        #     'automatic': True,
        #     'explanation': "bw2data 4.0 release requires all database be reprocessed"
        # }
    }

    @classmethod
    def explain(cls, key):
        return cls.UPDATES[key]["explanation"]

    @classmethod
    def do_update(cls, key):
        method = getattr(cls, cls.UPDATES[key]["method"])
        method()
        preferences["updates"][key] = True
        preferences.flush()

    @classmethod
    def check_status(cls, verbose=True):
        """Check if updates need to be applied.

        Returns:
            List of needed updates (strings), if any.

        """
        cls.set_initial_updates()
        updates = sorted(
            [
                key
                for key in cls.UPDATES
                if not preferences["updates"].get(key)
                and not cls.UPDATES[key]["automatic"]
            ]
        )
        if updates and verbose:
            warnings.warn(UPTODATE_WARNING)
        return updates

    @classmethod
    def set_initial_updates(cls):
        if "updates" in preferences:
            return
        SQL = "PRAGMA table_info(activitydataset)"
        with sqlite3.connect(sqlite3_lci_db.db.database) as conn:
            column_names = {x[1] for x in conn.execute(SQL)}
        if "code" in column_names:
            preferences["updates"] = {key: True for key in cls.UPDATES}
        else:
            preferences["updates"] = {}

    @classmethod
    def check_automatic_updates(cls):
        """Get list of automatic updates to be applied"""
        cls.set_initial_updates()
        return sorted(
            [
                key
                for key in cls.UPDATES
                if not preferences["updates"].get(key) and cls.UPDATES[key]["automatic"]
            ]
        )

    @classmethod
    def reprocess_all_1_0(cls):
        """1.0: Reprocess all to make sure default 'loc' value inserted when not specified."""
        cls._reprocess_all()

    @classmethod
    def schema_change_20_compound_keys(cls):
        with sqlite3.connect(sqlite3_lci_db.db.database) as conn:
            print("Update ActivityDataset table schema and data")
            conn.executescript(UPDATE_ACTIVITYDATASET)
            print("Updating ExchangeDataset table schema and data")
            conn.executescript(UPDATE_EXCHANGEDATASET)
            print("Finished with schema change")

    @classmethod
    def database_search_directories_20(cls):
        shutil.rmtree(projects.request_directory("whoosh"))
        for db in databases:
            if databases[db].get("searchable"):
                databases[db]["searchable"] = False
                print("Reindexing database {}".format(db))
                Database(db).make_searchable()

    @classmethod
    def processed_data_format_change_23(cls):
        processed_dir = projects.dir / "processed"
        for filename in os.listdir(processed_dir):
            fp = processed_dir / filename
            if fp.is_dir():
                continue
            if not filename.endswith(".pickle"):
                continue
            np.save(fp[:-7] + ".npy", pickle.load(open(fp, "rb")), allow_pickle=False)

    @classmethod
    def expire_all_processed_data_40(cls):
        cls._reprocess_all()

    @classmethod
    def _reprocess_all(cls):
        objects = [
            (methods, Method, "LCIA methods"),
            (weightings, Weighting, "LCIA weightings"),
            (normalizations, Normalization, "LCIA normalizations"),
            (databases, Database, "LCI databases"),
        ]

        for (meta, klass, name) in objects:
            if meta.list:
                print("Updating all %s" % name)

                pbar = pyprind.ProgBar(
                    len(meta), title="Brightway2 {} objects:".format(name), monitor=True
                )

                for index, key in enumerate(meta):
                    obj = klass(key)
                    obj.process()
                    # Free memory
                    obj = None

                    pbar.update()
                print(pbar)



from .backends.schema import get_id
from .ia_data_store import ImpactAssessmentDataStore
from .meta import weightings, normalizations
from .project import writable_project
from .utils import as_uncertainty_dict
from .validate import weighting_validator, normalization_validator


class Weighting(ImpactAssessmentDataStore):
    """
    LCIA weighting data - used to combine or compare different impact categories.

    The data schema for weighting is a one-element list:

    .. code-block:: python

            Schema(All(
                [uncertainty_dict],
                Length(min=1, max=1)
            ))

    """

    _metadata = weightings
    validator = weighting_validator
    matrix = "weighting_matrix"

    @writable_project
    def write(self, data):
        """Because of DataStore assumptions, need a one-element list"""
        if self.name not in self._metadata:
            self.register()
        if not isinstance(data, list) or not len(data) == 1:
            raise ValueError("Weighting data must be one-element list")
        super(Weighting, self).write(data)

    def process_row(self, row):
        """Return an empty tuple (as ``dtype_fields`` is empty), and the weighting uncertainty dictionary."""
        return {**as_uncertainty_dict(row), "row": 0}


class Normalization(ImpactAssessmentDataStore):
    """
    LCIA normalization data - used to transform meaningful units, like mass or damage, into "person-equivalents" or some such thing.

    The data schema for IA normalization is:

    .. code-block:: python

            Schema([
                [valid_tuple, maybe_uncertainty]
            ])

    where:
        * ``valid_tuple`` is a dataset identifier, like ``("biosphere", "CO2")``
        * ``maybe_uncertainty`` is either a number or an uncertainty dictionary

    """

    _metadata = normalizations
    validator = normalization_validator
    matrix = "normalization_matrix"

    def process_row(self, row):
        """Given ``(flow key, amount)``, return a dictionary for array insertion."""
        return {
            **as_uncertainty_dict(row[1]),
            "row": get_id(row[0]),
        }



from whoosh.fields import TEXT, ID, Schema

bw2_schema = Schema(
    name=TEXT(stored=True, sortable=True),
    comment=TEXT(stored=True),
    product=TEXT(stored=True, sortable=True),
    categories=TEXT(stored=True),
    synonyms=TEXT(stored=True),
    location=TEXT(stored=True, sortable=True),
    database=TEXT(stored=True),
    code=ID(unique=True, stored=True),
)



from bw2data.tests import BW2DataTest
from bw2data import Updates, config
import random


class UpdatesTest(BW2DataTest):
    def test_set_updates_clean_install(self):
        self.assertFalse("updates" in config.p)
        self.assertFalse(Updates.check_status())
        self.assertEqual(len(config.p["updates"]), len(Updates.UPDATES))

    def test_explain(self):
        key = random.choice(list(Updates.UPDATES.keys()))
        self.assertEqual(Updates.UPDATES[key]["explanation"], Updates.explain(key))

    def test_do_updates(self):
        # Test with mock that overwrites UPDATES?
        pass



from bw2data import databases, methods, geomapping, projects
from bw2data.tests import bw2test


@bw2test
def test_bw2test_decorator():
    assert list(databases) == []
    assert list(methods) == []
    assert len(geomapping) == 1
    assert "GLO" in geomapping
    assert len(projects) == 2
    assert "default" in projects



from .data_store import ProcessedDataStore
from bw_processing import safe_filename
import hashlib
import string


def abbreviate(names, length=8):
    """Take a tuple or list, and construct a string, doing the following:

    First, apply :func:`.filesystem.safe_filename` to each element in ``names``.

    Next, take the following, in order:
        * The first word of the first element in names, lower-cased, where word is defined as everything up to the first empty space character.
        * Join the rest of the first element (i.e. after the first word) with all other elements. Use the empty space character to join.
        * In this long string separated by spaces, take the lowercase first character of each word. Add the first word to this new string.
        * Finally, add a dash, and then the MD5 hash of the entire identifier, where each element is joined by a dash character.

    ``('ReCiPe Endpoint (E,A)', 'human health', 'ionising radiation')`` becomes ``'recipee(hhir-70eeef20a20deb6347ad428e3f6c5f3c'``.

    The MD5 hash is needed because taking the first characters doesn't guarantee unique strings.

    """
    safe_names = [safe_filename(x, False) for x in names]
    abbrev = lambda x: x if x[0] in string.digits else x[0].lower()
    name = " ".join(safe_names).split(" ")[0].lower() + "".join(
        [abbrev(x) for x in " ".join(safe_names).split(" ")[1:]]
    )
    return name + "." + str(hashlib.md5(("-".join(names)).encode("utf-8")).hexdigest())


class ImpactAssessmentDataStore(ProcessedDataStore):
    """
A subclass of ``DataStore`` for impact assessment methods.

IA objects are hierarchically structured, and their identifier uses this structure, like ``('ecological scarcity 2006', 'total', 'natural resources')``. The identifier must be a ``tuple``, i.e. ``()``, not a ``list``, i.e. ``[]``. The identifier should only contain unicode strings, and can be of any length >= 1.

Because impact assessment methods are identified by a tuple of strings, e.g. ``('ReCiPe Endpoint (E,A)', 'human health', 'ionising radiation')``, we need to transform this identifier before it can be used e.g. as a filename. We do this using the :func:`.abbreviate` function, which returns a single unicode string.

Args:
    * *name* (tuple): Name of the IA object to manage. Must be a tuple of unicode strings.

    """

    def __str__(self):
        return "Brightway2 %s: %s" % (self.__class__.__name__, ": ".join(self.name))

    def get_abbreviation(self):
        """Retrieve the abbreviation of the method identifier from the metadata store. See class documentation."""
        self.register()
        return self.metadata["abbreviation"]

    def copy(self, name=None):
        """Make a copy of the method, including its CFs and metadata.

        If ``name`` is not provided, add "Copy of" to the last element of the original name, e.g. ``("foo", "bar")`` becomes ``("foo", "Copy of bar")``

        Args:
            * *name* (tuple, optional): Name of the new method.

        Returns:
            The new object.

        """
        if name is None:
            name = self.name[:-1] + ("Copy of " + self.name[-1],)
        else:
            name = tuple(name)

        return super(ImpactAssessmentDataStore, self).copy(name)

    def register(self, **kwargs):
        """Register an object with the metadata store.

        The metadata key ``abbreviation`` is set automatically.

        Objects must be registered before data can be written. If this object is not yet registered in the metadata store, a warning is written to **stdout**.

        Takes any number of keyword arguments.

        """
        kwargs["abbreviation"] = abbreviate(self.name)
        super(ImpactAssessmentDataStore, self).register(**kwargs)

    @property
    def filename(self):
        return self.get_abbreviation()



from . import projects
from .errors import PickleError
from .fatomic import open as atomic_open
from .project import writable_project
from .utils import maybe_path
from collections.abc import MutableMapping
from time import time
import bz2
import os
import pickle
import random

try:
    import anyjson
except ImportError:
    anyjson = None
    import json


class JsonWrapper(object):
    @classmethod
    def dump(self, data, filepath):
        with atomic_open(filepath, "w") as f:
            if anyjson:
                f.write(anyjson.serialize(data))
            else:
                json.dump(data, f, indent=2)

    @classmethod
    def dump_bz2(self, data, filepath):
        with atomic_open(filepath, "wb") as f:
            with bz2.BZ2File(f.name, "wb") as b:
                b.write((JsonWrapper.dumps(data)).encode("utf-8"))

    @classmethod
    def load(self, file):
        if anyjson:
            return anyjson.deserialize(open(file, encoding="utf-8").read())
        else:
            return json.load(open(file, encoding="utf-8"))

    @classmethod
    def load_bz2(self, filepath):
        return JsonWrapper.loads((bz2.BZ2File(filepath).read()).decode("utf-8"))

    @classmethod
    def dumps(self, data):
        if anyjson:
            return anyjson.serialize(data)
        else:
            return json.dumps(data)

    @classmethod
    def loads(self, data):
        if anyjson:
            return anyjson.deserialize(data)
        else:
            return json.loads(data)


class JsonSanitizer(object):
    @classmethod
    def sanitize(cls, data):
        if isinstance(data, tuple):
            return {"__tuple__": True, "data": [cls.sanitize(x) for x in data]}
        elif isinstance(data, dict):
            return {
                "__dict__": True,
                "keys": [cls.sanitize(x) for x in data.keys()],
                "values": [cls.sanitize(x) for x in data.values()],
            }
        elif isinstance(data, list):
            return [cls.sanitize(x) for x in data]
        else:
            return data

    @classmethod
    def load(cls, data):
        if isinstance(data, dict):
            if "__tuple__" in data:
                return tuple([cls.load(x) for x in data["data"]])
            elif "__dict__" in data:
                return dict(
                    zip(
                        [cls.load(x) for x in data["keys"]],
                        [cls.load(x) for x in data["values"]],
                    )
                )
            else:
                raise ValueError
        elif isinstance(data, list):
            return [cls.load(x) for x in data]
        else:
            return data


class SerializedDict(MutableMapping):
    """Base class for dictionary that can be `serialized <http://en.wikipedia.org/wiki/Serialization>`_ to or unserialized from disk. Uses JSON as its storage format. Has most of the methods of a dictionary.

    Upon instantiation, the serialized dictionary is read from disk."""

    def __init__(self, dirpath=None):
        if not getattr(self, "filename"):
            raise NotImplementedError(
                "SerializedDict must be subclassed, and the filename must be set."
            )
        self.filepath = (maybe_path(dirpath) or projects.dir) / self.filename
        self.load()

    def load(self):
        """Load the serialized data. Creates the file if not yet present."""
        try:
            self.data = self.deserialize()
        except IOError:
            # Create if not present
            self.data = {}
            self.flush()

    def flush(self):
        """Serialize the current data to disk."""
        self.serialize()

    @property
    def list(self):
        """List the keys of the dictionary. This is a property, and does not need to be called."""
        return sorted(self.data.keys())

    def __getitem__(self, key):
        if isinstance(key, list):
            key = tuple(key)
        return self.data[key]

    @writable_project
    def __setitem__(self, key, value):
        self.data[key] = value
        self.flush()

    def __contains__(self, key):
        return key in self.data

    def __str__(self):
        if not len(self):
            return "{} dictionary with 0 objects".format(self.__class__.__name__)
        elif len(self) > 20:
            return (
                "{} dictionary with {} objects, including:"
                "{}\nUse `list(this object)` to get the complete list."
            ).format(
                self.__class__.__name__,
                len(self),
                "".join(["\n\t{}".format(x) for x in sorted(self.data)[:10]]),
            )
        else:
            return ("{} dictionary with {} object(s):{}").format(
                self.__class__.__name__,
                len(self),
                "".join(["\n\t{}".format(x) for x in sorted(self.data)]),
            )

    __repr__ = lambda x: str(x)

    def __delitem__(self, name):
        del self.data[name]
        self.flush()

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __hash__(self):
        return hash(self.data)

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    @writable_project
    def serialize(self, filepath=None):
        """Method to do the actual serialization. Can be replaced with other serialization formats.

        Args:
            * *filepath* (str, optional): Provide an alternate filepath (e.g. for backup).

        """
        with atomic_open(filepath or self.filepath, "w") as f:
            f.write(JsonWrapper.dumps(self.pack(self.data)))

    def deserialize(self):
        """Load the serialized data. Can be replaced with other serialization formats."""
        return self.unpack(JsonWrapper.load(self.filepath))

    def pack(self, data):
        """Transform the data, if necessary. Needed because JSON must have strings as dictionary keys."""
        return data

    def unpack(self, data):
        """Return serialized data to true form."""
        return data

    def random(self):
        """Return a random key."""
        if not self.data:
            return None
        else:
            return random.choice(list(self.data.keys()))

    def backup(self):
        """Write a backup version of the data to the ``backups`` directory."""
        filepath = os.path.join(
            projects.dir, "backups", self.filename + ".%s.backup" % int(time())
        )
        self.serialize(filepath)


class PickledDict(SerializedDict):
    """Subclass of ``SerializedDict`` that uses the pickle format instead of JSON."""

    @writable_project
    def serialize(self):
        with atomic_open(self.filepath, "wb") as f:
            pickle.dump(self.pack(self.data), f, protocol=4)

    def deserialize(self):
        try:
            return self.unpack(pickle.load(open(self.filepath, "rb")))
        except ImportError:
            TEXT = "Pickle deserialization error in file '%s'" % self.filepath
            raise PickleError(TEXT)


class CompoundJSONDict(SerializedDict):
    """Subclass of ``SerializedDict`` that allows tuples as dictionary keys (not allowed in JSON)."""

    def pack(self, data):
        """Transform the dictionary to a list because JSON can't handle lists as keys"""
        return [(k, v) for k, v in data.items()]

    def unpack(self, data):
        """Transform data back to a dictionary"""
        return dict([(tuple(x[0]), x[1]) for x in data])



from . import databases
from .errors import InvalidExchange
from .utils import get_activity
from collections.abc import MutableMapping
from numbers import Number
from stats_arrays import uncertainty_choices


class ProxyBase(MutableMapping):
    def __init__(self, data, *args, **kwargs):
        self._data = data

    def as_dict(self):
        return self._data

    def __str__(self):
        return "Instance of base proxy class"

    __repr__ = lambda x: str(x)

    def __contains__(self, key):
        return key in self._data

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __eq__(self, other):
        return self._dict == other

    def __hash__(self):
        return hash(self._dict)


class ActivityProxyBase(ProxyBase):
    def __str__(self):
        if self.valid():
            return "'{}' ({}, {}, {})".format(
                self.get("name"),
                self.get("unit"),
                self.get("location"),
                self.get("categories"),
            )
        else:
            return "Activity with missing fields (call ``valid(why=True)`` to see more)"

    @property
    def key(self):
        return (self.get("database"), self.get("code"))

    def __eq__(self, other):
        return self.key == other

    def __lt__(self, other):
        if not isinstance(other, ActivityProxyBase):
            raise TypeError
        else:
            return self.key < other.key

    def __hash__(self):
        return hash(self.key)

    def __getitem__(self, key):
        # Basically a hack to let this act like a tuple with two
        # elements, database and code. Useful for using as functional unit in LCA.
        if key == 0:
            return self["database"]
        elif key == 1:
            return self["code"]
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def valid(self, why=False):
        errors = []
        if not self.get("database"):
            errors.append("Missing field ``database``")
        elif self.get("database") not in databases:
            errors.append("``database`` refers to unknown database")
        if not self.get("code"):
            errors.append("Missing field ``code``")
        if not self.get("name"):
            errors.append("Missing field ``name``")
        if errors:
            if why:
                return (False, errors)
            else:
                return False
        else:
            return True

    def lca(self, method=None, amount=1.0):
        """Shortcut to construct an LCA object for this activity."""
        from bw2calc import LCA

        lca = LCA({self: amount}, method=method)
        lca.lci()
        if method is not None:
            lca.lcia()
        lca.fix_dictionaries()
        return lca


class ExchangeProxyBase(ProxyBase):
    def __str__(self):
        if self.valid():
            return "Exchange: {} {} {} to {}>".format(
                self.amount, self.unit, self.input, self.output
            )
        else:
            return "Exchange with missing fields (call ``valid(why=True)`` to see more)"

    def __lt__(self, other):
        if not isinstance(other, ExchangeProxyBase):
            raise TypeError
        else:
            return (self.input.key, self.output.key) < (
                other.input.key,
                other.output.key,
            )

    def _get_input(self):
        """Get or set the exchange input.

        When getting, returns an `Activity` - this will raise an error if the linked activity doesn't yet exist.

        When setting, either an `Activity` or a tuple can be given. The linked activity does not have to exist yet."""
        if not self.get("input"):
            raise InvalidExchange("Missing valid data for `input` field")
        elif not hasattr(self, "_input"):
            self._input = get_activity(self["input"])
        return self._input

    def _set_input(self, value):
        if isinstance(value, ActivityProxyBase):
            self._input = value
            self._data["input"] = value.key
        elif isinstance(value, (tuple, list)):
            self._data["input"] = value
        else:
            raise ValueError("Provided input data is invalid")

    def _get_output(self):
        """Get or set the exchange output.

        When getting, returns an `Activity` - this will raise an error if the linked activity doesn't yet exist.

        When setting, either an `Activity` or a tuple can be given. The linked activity does not have to exist yet."""
        if not self.get("output"):
            raise InvalidExchange("Missing valid data for `output` field")
        elif not hasattr(self, "_output"):
            self._output = get_activity(self["output"])
        return self._output

    def _set_output(self, value):
        if isinstance(value, ActivityProxyBase):
            self._output = value
            self._data["output"] = value.key
        elif isinstance(value, (tuple, list)):
            self._data["output"] = value
        else:
            raise ValueError("Provided input data is invalid")

    input = property(_get_input, _set_input)
    output = property(_get_output, _set_output)

    def __setitem__(self, key, value):
        if key == "input":
            self.input = value
        elif key == "output":
            self.output = value
        else:
            self._data[key] = value

    def valid(self, why=False):
        errors = []
        if not self.get("input"):
            errors.append("Missing field ``input``")
        elif not isinstance(self["input"], tuple):
            errors.append("Field ``input`` must be a tuple")
        elif self["input"][0] not in databases:
            errors.append(
                "Input database ``{}`` doesn't exist".format(self["input"][0])
            )

        if not self.get("output"):
            errors.append("Missing field ``output``")
        elif not isinstance(self["output"], tuple):
            errors.append("Field ``output`` must be a tuple")
        elif self["output"][0] not in databases:
            errors.append(
                "Output database ``{}`` doesn't exist".format(self["output"][0])
            )

        if not isinstance(self.get("amount", None), Number):
            errors.append("Invalid or missing field ``amount``")
        if not self.get("type"):
            errors.append("Missing field ``type``")
        if errors:
            if why:
                return (False, errors)
            else:
                return False
        else:
            return True

    @property
    def unit(self):
        """Get exchange unit.

        Separate property because the unit is a property of the input, not the exchange itself."""
        return self.input.get("unit")

    @property
    def amount(self):
        return self.get("amount")

    @property
    def uncertainty(self):
        """Get uncertainty dictionary that can be used in uncertainty analysis."""
        KEYS = {
            "uncertainty type",
            "loc",
            "scale",
            "shape",
            "minimum",
            "maximum",
            "negative",
        }
        return {k: v for k, v in self.items() if k in KEYS}

    @property
    def uncertainty_type(self):
        """Get uncertainty type as a ``stats_arrays`` class."""
        return uncertainty_choices[self.get("uncertainty type", 0)]

    def random_sample(self, n=100):
        """Draw a random sample from this exchange."""
        ut = self.uncertainty_type
        array = ut.from_dicts(self.uncertainty)
        return ut.bounded_random_variables(array, n).ravel()

    def lca(self, method=None, amount=None):
        """Shortcut to construct an LCA object for this exchange **input**.

        Uses the exchange amount if no other amount is provided."""
        from bw2calc import LCA

        if amount is None:
            amount = self["amount"]

        lca = LCA({self.input: amount}, method=method)
        lca.lci()
        if method is not None:
            lca.lcia()
        lca.fix_dictionaries()
        return lca



from bw2data.tests import bw2test
from bw2data import parameters, Database, get_activity
from bw2data.parameters import (
    ActivityParameter,
    DatabaseParameter,
    Group,
    GroupDependency,
    ParameterizedExchange,
    parameters,
    ProjectParameter,
)
from bw2parameters.errors import MissingName
from peewee import IntegrityError
import pytest
import time
import re


# Regex to search for UUID: https://stackoverflow.com/a/18359032
uuid4hex = re.compile(
    r"[0-9a-f]{8}[0-9a-f]{4}4[0-9a-f]{3}[89ab][0-9a-f]{3}[0-9a-f]{12}", re.I
)

######################
### Project parameters
######################


@bw2test
def test_project_parameters():
    assert not len(parameters)
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    assert obj.name == "foo"
    assert obj.amount == 3.14
    assert obj.data == {"uncertainty type": 0}
    assert str(obj)
    assert isinstance(str(obj), str)


@bw2test
def test_project_parameter_autocreate_group():
    assert not Group.select().count()
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    assert Group.get(name="project")
    assert not Group.get(name="project").fresh


@bw2test
def test_expire_downstream():
    Group.create(fresh=True, name="A")
    Group.create(fresh=True, name="B")
    GroupDependency.create(group="B", depends="A")
    assert Group.get(name="A").fresh
    assert Group.get(name="B").fresh
    ProjectParameter.expire_downstream("A")
    assert not Group.get(name="B").fresh


@bw2test
def test_project_parameters_ordering():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    with pytest.raises(TypeError):
        obj < 0
    assert not (obj < obj)
    another = ProjectParameter.create(name="bar", formula="2 * foo",)
    assert another < obj


@bw2test
def test_project_parameters_dict():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    expected = {
        "name": "foo",
        "amount": 3.14,
        "uncertainty type": 0,
    }
    assert obj.dict == expected


@bw2test
def test_project_parameters_load():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    another = ProjectParameter.create(name="bar", formula="2 * foo",)
    expected = {
        "foo": {"amount": 3.14, "uncertainty type": 0},
        "bar": {"formula": "2 * foo"},
    }
    assert ProjectParameter.load() == expected
    assert ProjectParameter.load("project") == expected
    assert ProjectParameter.load("foo") == expected


@bw2test
def test_project_parameters_static():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    another = ProjectParameter.create(name="bar", formula="2 * foo",)
    assert ProjectParameter.static() == {"foo": 3.14, "bar": None}
    assert ProjectParameter.static(only=["foo"]) == {"foo": 3.14}
    ProjectParameter.recalculate()
    assert ProjectParameter.static() == {"foo": 3.14, "bar": 2 * 3.14}
    assert ProjectParameter.static(only=["bar"]) == {"bar": 2 * 3.14}


@bw2test
def test_project_parameters_expired():
    assert not ProjectParameter.expired()
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    assert ProjectParameter.expired()
    ProjectParameter.recalculate()
    assert not ProjectParameter.expired()


@bw2test
def test_project_parameters_recalculate():
    ProjectParameter.recalculate()
    Group.create(name="project")
    ProjectParameter.recalculate()
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    another = ProjectParameter.create(name="bar", formula="2 * foo",)
    ProjectParameter.recalculate()
    obj = ProjectParameter.get(name="bar")
    assert obj.amount == 2 * 3.14


@bw2test
def test_project_parameters_expire_downstream():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    Group.create(name="bar")
    GroupDependency.create(group="bar", depends="project")
    assert Group.get(name="bar").fresh
    ProjectParameter.recalculate()
    assert not Group.get(name="bar").fresh


@bw2test
def test_project_autoupdate_triggers():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    first = Group.get(name="project").updated
    time.sleep(1.1)
    another = ProjectParameter.create(name="bar", formula="2 * foo",)
    second = Group.get(name="project").updated
    assert first != second
    time.sleep(1.1)
    ProjectParameter.update(amount=7).execute()
    third = Group.get(name="project").updated
    assert second != third
    time.sleep(1.1)
    ProjectParameter.get(name="foo").delete_instance()
    fourth = Group.get(name="project").updated
    assert fourth != third


@bw2test
def test_project_name_uniqueness():
    obj = ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    with pytest.raises(IntegrityError):
        ProjectParameter.create(
            name="foo", amount=7,
        )


@bw2test
def test_project_parameter_dependency_chain():
    ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    ProjectParameter.create(name="bar", amount=6.28, formula="foo * 2")
    expected = [
        {"kind": "project", "group": "project", "names": set(["foo"])},
    ]
    assert ProjectParameter.dependency_chain() == expected


@bw2test
def test_project_parameter_dependency_chain_missing():
    ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    ProjectParameter.create(name="baz", amount=8, formula="foo * bar")
    with pytest.raises(MissingName):
        ProjectParameter.dependency_chain()


@bw2test
def test_project_parameter_depend_within_group():
    ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    ProjectParameter.create(name="baz", amount=8, formula="foo * 2")
    assert ProjectParameter.is_dependency_within_group("foo")
    assert not ProjectParameter.is_dependency_within_group("baz")


@bw2test
def test_project_parameter_is_deletable():
    """ Project parameters can be deleted if they are no dependencies.
    """
    ProjectParameter.create(name="foo", amount=3.14)
    assert ProjectParameter.get(name="foo").is_deletable()


@bw2test
def test_project_parameter_is_not_deletable_project():
    ProjectParameter.create(name="foo", amount=3.14)
    ProjectParameter.create(name="bar", amount=1, formula="foo * 2")
    assert not ProjectParameter.get(name="foo").is_deletable()


@bw2test
def test_project_parameter_is_not_deletable_database():
    ProjectParameter.create(name="foo", amount=3.14)
    Database("B").register()
    DatabaseParameter.create(database="B", name="bar", amount=1, formula="foo * 5")
    # Recalculate to build GroupDependencies.
    parameters.recalculate()
    assert not ProjectParameter.get(name="foo").is_deletable()


@bw2test
def test_project_parameter_is_not_deletable_activity():
    ProjectParameter.create(name="foo", amount=3.14)
    Database("B").register()
    ActivityParameter.create(
        group="baz", database="B", code="first", name="bar", amount=0, formula="foo + 4"
    )
    # Recalculate to build GroupDependencies.
    parameters.recalculate()
    assert not ProjectParameter.get(name="foo").is_deletable()


@bw2test
def test_project_parameter_formula_update():
    """ Update formulas only where the name of the parameter is an exact match.
    """
    ProjectParameter.create(name="foo", amount=3.14)
    ProjectParameter.create(name="foobar", amount=6.28)
    ProjectParameter.create(name="bar", amount=1, formula="foo + 2")
    ProjectParameter.create(name="baz", amount=1, formula="foobar * 3")
    assert (
        ProjectParameter.select()
        .where(ProjectParameter.formula.contains("foo"))
        .count()
        == 2
    )
    ProjectParameter.update_formula_parameter_name("foo", "efficiency")
    assert ProjectParameter.get(name="bar").formula == "efficiency + 2"
    assert ProjectParameter.get(name="baz").formula == "foobar * 3"


#######################
### Database parameters
#######################


@bw2test
def test_create_database_parameters():
    assert not len(parameters)
    obj = DatabaseParameter.create(database="bar", name="foo", amount=3.14,)
    assert obj.name == "foo"
    assert obj.database == "bar"
    assert obj.amount == 3.14
    assert str(obj)
    assert isinstance(str(obj), str)
    assert len(parameters)


@bw2test
def test_database_parameters_group_autocreated():
    assert not Group.select().count()
    obj = DatabaseParameter.create(database="bar", name="foo", amount=3.14,)
    assert Group.get(name="bar")
    assert not Group.get(name="bar").fresh


@bw2test
def test_database_parameters_expired():
    assert not DatabaseParameter.expired("bar")
    DatabaseParameter.create(
        database="bar", name="foo", amount=3.14,
    )
    assert DatabaseParameter.expired("bar")


@bw2test
def test_database_parameters_dict():
    obj = DatabaseParameter.create(database="bar", name="foo", amount=3.14,)
    expected = {
        "database": "bar",
        "name": "foo",
        "amount": 3.14,
    }
    assert obj.dict == expected


@bw2test
def test_database_parameters_load():
    DatabaseParameter.create(
        database="bar", name="foo", amount=3.14,
    )
    DatabaseParameter.create(database="bar", name="baz", formula="foo + baz")
    expected = {
        "foo": {"database": "bar", "amount": 3.14},
        "baz": {"database": "bar", "formula": "foo + baz"},
    }
    assert DatabaseParameter.load("bar") == expected


@bw2test
def test_database_parameters_static():
    DatabaseParameter.create(
        database="bar", name="foo", amount=3.14,
    )
    DatabaseParameter.create(database="bar", name="baz", amount=7, formula="foo + baz")
    expected = {"foo": 3.14, "baz": 7}
    assert DatabaseParameter.static("bar") == expected
    assert DatabaseParameter.static("bar", only=["baz"]) == {"baz": 7}


@bw2test
def test_database_parameters_check():
    with pytest.raises(IntegrityError):
        DatabaseParameter.create(
            database="project", name="foo", amount=3.14,
        )


@bw2test
def test_database_autoupdate_triggers():
    obj = DatabaseParameter.create(database="A", name="foo", amount=3.14,)
    first = Group.get(name="A").updated
    time.sleep(1.1)
    another = DatabaseParameter.create(database="A", name="bar", formula="2 * foo",)
    second = Group.get(name="A").updated
    assert first != second
    time.sleep(1.1)
    DatabaseParameter.update(amount=7).execute()
    third = Group.get(name="A").updated
    assert second != third
    time.sleep(1.1)
    DatabaseParameter.get(name="foo").delete_instance()
    fourth = Group.get(name="A").updated
    assert fourth != third


@bw2test
def test_database_uniqueness_constraint():
    DatabaseParameter.create(
        database="A", name="foo", amount=3.14, data={"uncertainty type": 0}
    )
    with pytest.raises(IntegrityError):
        DatabaseParameter.create(
            database="A", name="foo", amount=7,
        )


@bw2test
def test_database_parameter_cross_database_constraint():
    """Database parameters cannot use parameters on other databases."""
    Database("B").register()
    Database("C").register()
    DatabaseParameter.create(
        database="B", name="car", amount=8,
    )
    DatabaseParameter.create(
        database="C", name="plane", formula="car ** 5",
    )
    with pytest.raises(MissingName):
        DatabaseParameter.recalculate("C")


@bw2test
def test_update_database_parameters():
    assert not Group.select().count()
    assert not GroupDependency.select().count()

    DatabaseParameter.create(
        database="A", name="B", amount=5,
    )
    o = DatabaseParameter.create(database="A", name="C", formula="B * 2 + foo",)
    Group.create(name="Zed")
    GroupDependency.create(group="Zed", depends="A")
    assert Group.get(name="A")
    with pytest.raises(MissingName):
        DatabaseParameter.recalculate("A")
    o.formula = "B * 2"
    o.save()
    DatabaseParameter.recalculate("A")
    assert Group.get(name="A").fresh
    assert DatabaseParameter.get(name="C").amount == 10
    assert not Group.get(name="Zed").fresh

    o.formula = "B * 2 + foo + bar"
    o.save()
    ProjectParameter.create(name="foo", amount=3.14, data={"uncertainty type": 0})
    ProjectParameter.create(
        name="bar", formula="2 * foo",
    )
    assert Group.get(name="project")
    Database("A").register()

    obj = DatabaseParameter.get(name="C")
    assert obj.amount != 3.14 * 3 + 10
    with pytest.raises(GroupDependency.DoesNotExist):
        GroupDependency.get(group="A", depends="project")

    DatabaseParameter.recalculate("A")
    assert GroupDependency.get(group="A", depends="project")
    assert Group.get(name="A")
    assert Group.get(name="project")
    obj = DatabaseParameter.get(name="C")
    assert obj.amount == 3.14 * 3 + 10


@bw2test
def test_database_parameter_dependency_chain():
    Database("B").register()
    DatabaseParameter.create(
        database="B", name="car", formula="2 ** fly", amount=8,
    )
    DatabaseParameter.create(
        database="B", name="bike", formula="car - hike", amount=2,
    )
    ProjectParameter.create(
        name="hike", formula="2 * 2 * 2", amount=6,
    )
    ProjectParameter.create(
        name="fly", formula="3", amount=3,
    )
    expected = [
        {"kind": "project", "group": "project", "names": set(["fly", "hike"])},
    ]
    assert DatabaseParameter.dependency_chain("B") == expected
    assert DatabaseParameter.dependency_chain("missing") == []


@bw2test
def test_database_parameter_dependency_chain_missing():
    Database("B").register()
    DatabaseParameter.create(
        database="B", name="car", formula="2 ** fly", amount=8,
    )
    ProjectParameter.create(
        name="hike", formula="2 * 2 * 2", amount=6,
    )
    with pytest.raises(MissingName):
        DatabaseParameter.dependency_chain("B")


@bw2test
def test_database_parameter_dependency_chain_include_self():
    Database("B").register()
    DatabaseParameter.create(
        database="B", name="car", formula="2 ** fly", amount=8,
    )
    DatabaseParameter.create(
        database="B", name="truck", formula="car * 5",
    )
    ProjectParameter.create(
        name="fly", formula="3", amount=3,
    )
    expected = [
        {"kind": "project", "group": "project", "names": set(["fly"])},
        {"kind": "database", "group": "B", "names": set(["car"])},
    ]
    # Method now also includes required names within group
    assert DatabaseParameter.dependency_chain("B", include_self=True) == expected


@bw2test
def test_database_parameter_depend_within_group():
    Database("B").register()
    Database("C").register()
    DatabaseParameter.create(
        database="B", name="car", formula="2 ** fly", amount=8,
    )
    DatabaseParameter.create(
        database="B", name="truck", amount=2, formula="car * 5",
    )
    DatabaseParameter.create(
        database="C", name="fly", amount=7,
    )
    DatabaseParameter.create(
        database="C", name="parade", amount=1, formula="fly * 2.7",
    )
    ProjectParameter.create(
        name="fly", formula="3", amount=3,
    )
    parameters.recalculate()
    assert DatabaseParameter.is_dependency_within_group("car", "B")
    assert not DatabaseParameter.is_dependency_within_group("truck", "B")
    assert DatabaseParameter.is_dependency_within_group("fly", "C")


@bw2test
def test_database_parameter_is_deletable():
    """ Database parameters can be deleted if they are no dependencies.
    """
    Database("B").register()
    DatabaseParameter.create(database="B", name="car", amount=8)
    assert DatabaseParameter.get(name="car").is_deletable()


@bw2test
def test_database_parameter_is_not_deletable_database():
    Database("B").register()
    DatabaseParameter.create(database="B", name="car", amount=8)
    DatabaseParameter.create(database="B", name="truck", formula="car * 5", amount=4)
    assert not DatabaseParameter.get(name="car").is_deletable()


@bw2test
def test_database_parameter_is_not_deletable_activity():
    Database("B").register()
    DatabaseParameter.create(database="B", name="car", amount=8)
    ActivityParameter.create(
        group="cars",
        database="B",
        code="first",
        name="bar",
        amount=0,
        formula="car + 4",
    )
    # Build GroupDependencies
    parameters.recalculate()
    assert not DatabaseParameter.get(name="car").is_deletable()


@bw2test
def test_database_parameter_is_dependent_on():
    """ Databases parameters can be dependent on project parameters.
    """
    Database("B").register()
    ProjectParameter.create(name="foo", amount=2)
    ProjectParameter.create(name="bar", amount=5)
    DatabaseParameter.create(database="B", name="baz", amount=1, formula="foo + 2")
    parameters.recalculate()
    assert DatabaseParameter.is_dependent_on("foo")
    assert not DatabaseParameter.is_dependent_on("bar")


@bw2test
def test_database_parameter_formula_update_project():
    """ Update formulas of database parameters, only update the formulas
    where the actual ProjectParameter is referenced.
    """
    ProjectParameter.create(name="foo", amount=2)
    ProjectParameter.create(name="tracks", amount=14)
    Database("B").register()
    Database("C").register()
    DatabaseParameter.create(database="B", name="bar", amount=1, formula="foo + 2")
    DatabaseParameter.create(database="C", name="bing", amount=1, formula="foo + 2")
    DatabaseParameter.create(database="C", name="foo", amount=8, formula="tracks * 2")
    parameters.recalculate()
    assert DatabaseParameter.get(name="bar").formula == "foo + 2"
    assert DatabaseParameter.get(name="bing").formula == "foo + 2"
    DatabaseParameter.update_formula_project_parameter_name("foo", "baz")
    assert DatabaseParameter.get(name="bar").formula == "baz + 2"
    assert DatabaseParameter.get(name="bing").formula == "foo + 2"


@bw2test
def test_database_parameter_formula_update_database():
    """ Update formulas of database parameters, only update the formulas
    where the actual DatabaseParameter is referenced.
    """
    ProjectParameter.create(name="foo", amount=2)
    Database("B").register()
    Database("C").register()
    DatabaseParameter.create(database="B", name="bar", amount=1, formula="foo + 2")
    DatabaseParameter.create(database="C", name="bing", amount=1, formula="foo + 2")
    DatabaseParameter.create(
        database="C", name="foo", amount=8,
    )
    parameters.recalculate()
    assert DatabaseParameter.get(name="bar").formula == "foo + 2"
    assert DatabaseParameter.get(name="bing").formula == "foo + 2"
    DatabaseParameter.update_formula_database_parameter_name("foo", "baz")
    assert DatabaseParameter.get(name="bar").formula == "foo + 2"
    assert DatabaseParameter.get(name="bing").formula == "baz + 2"


###########################
### Parameterized exchanges
###########################


@bw2test
def test_create_parameterized_exchange_missing_group():
    with pytest.raises(IntegrityError):
        obj = ParameterizedExchange.create(group="A", exchange=42, formula="foo + bar")


@bw2test
def test_create_parameterized_exchange():
    assert not ParameterizedExchange.select().count()
    ActivityParameter.insert_dummy("A", ("b", "c"))
    obj = ParameterizedExchange.create(group="A", exchange=42, formula="foo + bar")
    assert obj.group == "A"
    assert obj.exchange == 42
    assert obj.formula == "foo + bar"
    assert ParameterizedExchange.select().count()


@bw2test
def test_create_parameterized_exchange_nonunique():
    ActivityParameter.insert_dummy("A", ("b", "c"))
    ParameterizedExchange.create(group="A", exchange=42, formula="foo + bar")
    with pytest.raises(IntegrityError):
        ParameterizedExchange.create(group="B", exchange=42, formula="2 + 3")


#######################
### Activity parameters
#######################


@pytest.fixture
@bw2test
def chain():
    Database("B").register()
    Database("K").register()
    Group.create(name="G", order=["A"])
    ActivityParameter.create(
        group="A", database="B", code="C", name="D", formula="2 ** 3", amount=1,
    )
    ActivityParameter.create(
        group="A", database="B", code="E", name="F", formula="foo + bar + D", amount=2,
    )
    ActivityParameter.create(
        group="G", database="K", code="H", name="J", formula="F + D * 2", amount=3,
    )
    DatabaseParameter.create(
        database="B", name="foo", formula="2 ** 2", amount=5,
    )
    ProjectParameter.create(
        name="bar", formula="2 * 2 * 2", amount=6,
    )


@bw2test
def test_create_activity_parameter():
    assert not ActivityParameter.select().count()
    obj = ActivityParameter.create(
        group="A", database="B", code="C", name="D", amount=3.14
    )
    assert obj.group == "A"
    assert obj.database == "B"
    assert obj.code == "C"
    assert obj.name == "D"
    assert obj.amount == 3.14
    assert str(obj)
    assert isinstance(str(obj), str)
    assert ActivityParameter.select().count()
    assert len(parameters)


@bw2test
def test_activity_parameters_group_autocreated():
    assert not Group.select().count()
    ActivityParameter.create(group="A", database="B", code="C", name="D", amount=3.14)
    assert Group.get(name="A")
    assert not Group.get(name="A").fresh


@bw2test
def test_activity_parameter_expired():
    assert not ActivityParameter.expired("A")
    ActivityParameter.create(group="A", database="B", code="C", name="D", amount=3.14)
    assert ActivityParameter.expired("A")
    Group.get(name="A").freshen()
    assert not ActivityParameter.expired("A")


@bw2test
def test_activity_parameter_dict():
    a = ActivityParameter.create(
        group="A", database="B", code="C", name="D", amount=3.14
    )
    expected = {"database": "B", "code": "C", "name": "D", "amount": 3.14}
    assert a.dict == expected
    b = ActivityParameter.create(
        group="A",
        database="B",
        code="E",
        name="F",
        amount=7,
        data={"foo": "bar"},
        formula="7 * 1",
    )
    expected = {
        "database": "B",
        "code": "E",
        "name": "F",
        "amount": 7,
        "foo": "bar",
        "formula": "7 * 1",
    }
    assert b.dict == expected


@bw2test
def test_activity_parameter_load():
    ActivityParameter.create(
        group="A",
        database="B",
        code="E",
        name="F",
        amount=7,
        data={"foo": "bar"},
        formula="7 * 1",
    )
    expected = {
        "F": {
            "database": "B",
            "code": "E",
            "amount": 7,
            "foo": "bar",
            "formula": "7 * 1",
        }
    }
    assert ActivityParameter.load("A") == expected


def test_activity_parameter_static(chain):
    expected = {"D": 1, "F": 2}
    assert ActivityParameter.static("A") == expected
    expected = {}
    assert ActivityParameter.static("A", only=[]) == expected
    expected = {"D": 1}
    assert ActivityParameter.static("A", only=["D"]) == expected
    expected = {"D": 1, "F": 2, "foo": 5, "bar": 6}
    assert ActivityParameter.static("A", full=True) == expected
    expected = {"foo": 5, "bar": 6}
    assert ActivityParameter.static("A", full=True, only=["foo", "bar"]) == expected


@bw2test
def test_activity_parameter_recalculate_shortcut():
    assert not ActivityParameter.recalculate("A")
    ActivityParameter.create(group="A", database="B", code="C", name="D", amount=3.14)
    Group.get(name="A").freshen()
    assert not ActivityParameter.recalculate("A")


def test_activity_parameter_dependency_chain(chain):
    expected = [{"kind": "activity", "group": "A", "names": set(["D", "F"])}]
    assert ActivityParameter.dependency_chain("G") == expected
    expected = [
        {"kind": "database", "group": "B", "names": set(["foo"])},
        {"kind": "project", "group": "project", "names": set(["bar"])},
    ]
    assert ActivityParameter.dependency_chain("A") == expected


def test_activity_parameter_dependency_chain_missing(chain):
    """ Use unknown parameter 'K' in formula to test for MissingName error.
    """
    ActivityParameter.create(
        group="G", database="K", code="L", name="M", formula="foo + bar / K", amount=7,
    )
    with pytest.raises(MissingName):
        ActivityParameter.dependency_chain("G")


def test_activity_parameter_dependency_chain_includes_exchanges(chain):
    ProjectParameter.create(name="something_new", amount=10)
    db = Database("K")
    a = db.new_activity(code="something something danger zone", name="An activity")
    a.save()
    a.new_exchange(
        amount=0, input=a, type="production", formula="something_new + 4 - J"
    ).save()
    parameters.add_exchanges_to_group("G", a)

    expected = [
        {"kind": "activity", "group": "A", "names": {"D", "F"}},
        {"group": "project", "kind": "project", "names": {"something_new"}},
    ]
    assert ActivityParameter.dependency_chain("G") == expected


def test_activity_parameter_dependency_chain_include_self(chain):
    """ Out of the parameters 'D' and 'F' in group 'A', only 'D' counts
    as a dependency for group 'A'.

    This means that 'F' can be freely deleted, after which 'D' is no longer
    a dependency for group 'A' (as 'D' was a dependency of 'F') and can now
    also be deleted.
    """
    expected = [
        {"kind": "database", "group": "B", "names": set(["foo"])},
        {"kind": "project", "group": "project", "names": set(["bar"])},
    ]
    assert ActivityParameter.dependency_chain("A") == expected
    expected = [
        {"kind": "activity", "group": "A", "names": set(["D"])},
        {"kind": "database", "group": "B", "names": set(["foo"])},
        {"kind": "project", "group": "project", "names": set(["bar"])},
    ]
    assert ActivityParameter.dependency_chain("A", include_self=True) == expected


def test_activity_parameter_dependency_chain_include_self_exchanges(chain):
    """ Out of the parameters 'J' and 'H' in group 'G', only 'H' counts
    as a dependency as 'J' is not used by either 'H' or by any exchanges.
    """
    ActivityParameter.create(
        group="G", database="K", code="L", name="H", amount=7,
    )
    db = Database("K")
    a = db.new_activity(code="not a robot", name="actually an activity")
    a.save()
    a.new_exchange(amount=0, input=a, type="production", formula="15 / H").save()
    parameters.add_exchanges_to_group("G", a)

    expected = [
        {"kind": "activity", "group": "A", "names": set(["D", "F"])},
    ]
    assert ActivityParameter.dependency_chain("G") == expected
    expected = [
        {"kind": "activity", "group": "A", "names": set(["D", "F"])},
        {"kind": "activity", "group": "G", "names": set(["H"])},
    ]
    assert ActivityParameter.dependency_chain("G", include_self=True) == expected


def test_activity_parameter_depend_within_group(chain):
    """ When considering only dependencies within the given group. 'D' is
    a dependency within the group 'A', while 'F' is not.
    """
    assert ActivityParameter.is_dependency_within_group("D", "A")
    assert not ActivityParameter.is_dependency_within_group("F", "A")


def test_activity_parameter_depend_within_group_include(chain):
    """ The 'J' parameter in group 'G' depends on the 'F' parameter in group
    'A'. 'F' doesn't exist within the 'G' group but is instead linked to the
    'J' parameter through the 'G' group order.
    """
    parameters.recalculate()
    assert ActivityParameter.is_dependent_on("F", "A")
    assert not ActivityParameter.is_dependency_within_group("F", "G")
    assert ActivityParameter.is_dependency_within_group("F", "G", include_order=True)


@bw2test
def test_activity_parameter_dummy():
    assert not ActivityParameter.select().count()
    ActivityParameter.insert_dummy("A", ("B", "C"))
    assert ActivityParameter.select().count() == 1
    a = ActivityParameter.get()
    assert a.name.startswith("__dummy_") and uuid4hex.search(a.name)
    assert a.database == "B"
    assert a.amount == 0

    ActivityParameter.insert_dummy("A", ("B", "C"))
    assert ActivityParameter.select().count() == 1


@bw2test
def test_activity_parameter_multiple_dummies():
    assert not ActivityParameter.select().count()
    ActivityParameter.insert_dummy("A", ("B", "C"))
    ActivityParameter.insert_dummy("A", ("B", "D"))
    assert ActivityParameter.select().count() == 2
    assert all(
        ap.name.startswith("__dummy_") and uuid4hex.search(ap.name)
        for ap in ActivityParameter.select()
    )


def test_activity_parameter_static_dependencies(chain):
    expected = {"foo": 5, "bar": 6}
    assert ActivityParameter._static_dependencies("A") == expected
    expected = {"bar": 6, "D": 1, "F": 2}
    assert ActivityParameter._static_dependencies("G") == expected


@bw2test
def test_activity_parameter_recalculate_exchanges():
    db = Database("example")
    db.register()
    assert not len(parameters)
    assert not len(db)

    a = db.new_activity(code="A", name="An activity")
    a.save()
    b = db.new_activity(code="B", name="Another activity")
    b.save()
    a.new_exchange(
        amount=0, input=b, type="technosphere", formula="foo * bar + 4"
    ).save()

    project_data = [
        {"name": "foo", "formula": "green / 7",},
        {"name": "green", "amount": 7},
    ]
    parameters.new_project_parameters(project_data)

    database_data = [
        {"name": "red", "formula": "(foo + blue ** 2) / 5",},
        {"name": "blue", "amount": 12},
    ]
    parameters.new_database_parameters(database_data, "example")

    activity_data = [
        {
            "name": "reference_me",
            "formula": "sqrt(red - 20)",
            "database": "example",
            "code": "B",
        },
        {
            "name": "bar",
            "formula": "reference_me + 2",
            "database": "example",
            "code": "A",
        },
    ]
    parameters.new_activity_parameters(activity_data, "my group")

    parameters.add_exchanges_to_group("my group", a)
    ActivityParameter.recalculate_exchanges("my group")

    for exc in a.exchanges():
        # (((1 + 12 ** 2) / 5 - 20) ** 0.5 + 2) + 4
        assert exc.amount == 9


@bw2test
def test_pe_no_activities_parameter_group_error():
    db = Database("example")
    db.register()
    assert not len(parameters)
    assert not len(db)

    a = db.new_activity(code="A", name="An activity")
    a.save()
    a.new_exchange(amount=0, input=a, type="production").save()

    for exc in a.exchanges():
        obj = ParameterizedExchange(
            exchange=exc._document.id, group="my group", formula="1 + 1",
        )
        with pytest.raises(IntegrityError):
            obj.save()


@bw2test
def test_recalculate_exchanges_no_activities_parameters():
    db = Database("example")
    db.register()
    assert not len(parameters)
    assert not len(db)

    a = db.new_activity(code="A", name="An activity")
    a.save()
    a.new_exchange(amount=0, input=a, type="production", formula="foo + 4").save()

    project_data = [
        {"name": "foo", "formula": "green / 7",},
        {"name": "green", "amount": 7},
    ]
    parameters.new_project_parameters(project_data)

    assert ActivityParameter.select().count() == 0
    parameters.add_exchanges_to_group("my group", a)
    ActivityParameter.recalculate_exchanges("my group")

    for exc in a.exchanges():
        assert exc.amount == 5
        assert exc.get("formula")

    assert ActivityParameter.select().count() == 1
    a = ActivityParameter.get()
    assert a.name.startswith("__dummy_") and uuid4hex.search(a.name)


@bw2test
def test_activity_parameter_recalculate():
    Database("B").register()
    ActivityParameter.create(
        group="A", database="B", code="C", name="D", formula="2 ** 3"
    )
    ActivityParameter.create(
        group="A", database="B", code="E", name="F", formula="2 * D"
    )
    assert not Group.get(name="A").fresh
    ActivityParameter.recalculate("A")
    assert ActivityParameter.get(name="D").amount == 8
    assert ActivityParameter.get(name="F").amount == 16
    assert Group.get(name="A").fresh

    Database("K").register()
    ActivityParameter.create(
        group="G", database="K", code="H", name="J", formula="F + D * 2"
    )
    ActivityParameter.create(
        group="G", database="K", code="E", name="F", amount=3,
    )
    assert not Group.get(name="G").fresh
    with pytest.raises(MissingName):
        ActivityParameter.recalculate("G")

    assert not Group.get(name="G").fresh
    g = Group.get(name="G")
    g.order = ["A"]
    g.save()
    ActivityParameter.recalculate("G")
    assert Group.get(name="G").fresh
    assert ActivityParameter.get(name="J").amount == 19
    assert ActivityParameter.get(name="F", database="K").amount == 3

    DatabaseParameter.create(
        database="B", name="foo", formula="2 ** 2",
    )
    ProjectParameter.create(
        name="bar", formula="2 * 2 * 2",
    )
    a = ActivityParameter.get(database="B", code="E")
    a.formula = "foo + bar + D"
    a.save()
    assert not Group.get(name="A").fresh
    ActivityParameter.recalculate("A")
    assert ActivityParameter.get(database="B", code="E").amount == 4 + 8 + 8
    assert Group.get(name="A").fresh


def test_activity_parameter_is_deletable(chain):
    """ An activity parameter is deletable if it is not a dependency of another
    activity parameter.
    """
    # Ensure that all GroupDependencies exist.
    Group.get(name="G").expire()
    parameters.recalculate()

    # Is not used by any other activity parameter
    assert ActivityParameter.get(name="J", group="G").is_deletable()
    # Is used by the 'F' activity parameter from the same group.
    assert not ActivityParameter.get(name="D", group="A").is_deletable()
    # Is used by the 'J' activity parameter from group 'G'
    assert not ActivityParameter.get(name="F", group="A").is_deletable()


def test_activity_parameter_is_dependent_on(chain):
    """ An activity parameter can be dependent on any other type of parameter.
    """
    # Ensure that GroupDependencies exist.
    parameters.recalculate()

    # Some activity parameter is using the "bar" project parameter
    assert ActivityParameter.is_dependent_on("bar", "project")
    # Some activity parameter is using the "foo" database parameter
    assert ActivityParameter.is_dependent_on("foo", "B")
    # Some activity parameter is using the "F" activity parameter
    assert ActivityParameter.is_dependent_on("F", "A")
    # No activity parameter is dependent on the "J" activity parameter
    assert not ActivityParameter.is_dependent_on("J", "G")


def test_activity_parameter_formula_update_project(chain):
    ActivityParameter.create(
        group="G", database="K", code="AA", name="bar", amount=2,
    )
    ActivityParameter.create(
        group="G", database="K", code="BB", name="baz", amount=5, formula="bar * 6"
    )
    parameters.recalculate()
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + D"
    assert ActivityParameter.get(name="baz", group="G").formula == "bar * 6"
    ActivityParameter.update_formula_project_parameter_name("bar", "banana")
    assert ActivityParameter.get(name="F", group="A").formula == "foo + banana + D"
    assert ActivityParameter.get(name="baz", group="G").formula == "bar * 6"


def test_activity_parameter_formula_update_database(chain):
    ActivityParameter.create(
        group="G", database="K", code="AA", name="foo", amount=2,
    )
    ActivityParameter.create(
        group="G", database="K", code="BB", name="baz", amount=5, formula="foo * 6"
    )
    parameters.recalculate()
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + D"
    assert ActivityParameter.get(name="baz", group="G").formula == "foo * 6"
    ActivityParameter.update_formula_database_parameter_name("foo", "mango")
    assert ActivityParameter.get(name="F", group="A").formula == "mango + bar + D"
    assert ActivityParameter.get(name="baz", group="G").formula == "foo * 6"


def test_activity_parameter_formula_update_activity(chain):
    parameters.recalculate()
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + D"
    assert ActivityParameter.get(name="J", group="G").formula == "F + D * 2"
    ActivityParameter.update_formula_activity_parameter_name("D", "dingo")
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + dingo"
    assert ActivityParameter.get(name="J", group="G").formula == "F + D * 2"


def test_activity_parameter_formula_update_activity_include(chain):
    parameters.recalculate()
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + D"
    assert ActivityParameter.get(name="J", group="G").formula == "F + D * 2"
    ActivityParameter.update_formula_activity_parameter_name(
        "D", "dingo", include_order=True
    )
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + dingo"
    assert ActivityParameter.get(name="J", group="G").formula == "F + dingo * 2"


@bw2test
def test_activity_parameter_crossdatabase_triggers():
    ActivityParameter.create(group="A", database="B", name="C", code="D")
    with pytest.raises(IntegrityError):
        ActivityParameter.create(group="A", database="E", name="F", code="G")
    with pytest.raises(IntegrityError):
        a = ActivityParameter.get(name="C")
        a.database = "E"
        a.save()
    with pytest.raises(IntegrityError):
        ActivityParameter.update(database="C").execute()


@bw2test
def test_activity_parameter_crossgroup_triggers():
    ActivityParameter.create(
        group="A", database="B", name="C", code="D", amount=11,
    )
    with pytest.raises(IntegrityError):
        ActivityParameter.create(
            group="E", database="B", name="C", code="D", amount=1,
        )
    ActivityParameter.create(
        group="E", database="B", name="C", code="F", amount=1,
    )


@bw2test
def test_activity_parameter_autoupdate_triggers():
    obj = ActivityParameter.create(
        group="A", database="B", name="C", code="D", amount=11,
    )
    first = Group.get(name="A").updated
    time.sleep(1.1)
    another = ActivityParameter.create(
        group="A", database="B", code="E", name="F", formula="2 * foo",
    )
    second = Group.get(name="A").updated
    assert first != second
    time.sleep(1.1)
    ActivityParameter.update(amount=7).execute()
    third = Group.get(name="A").updated
    assert second != third
    time.sleep(1.1)
    ActivityParameter.get(name="F").delete_instance()
    fourth = Group.get(name="A").updated
    assert fourth != third


@bw2test
def test_activity_parameter_checks_uniqueness_constraints():
    ActivityParameter.create(
        group="A", database="B", name="C", code="D", amount=11,
    )
    with pytest.raises(IntegrityError):
        ActivityParameter.create(
            group="A", database="B", name="C", code="G", amount=111,
        )


@bw2test
def test_activity_parameter_checks():
    with pytest.raises(IntegrityError):
        ActivityParameter.create(group="project", database="E", name="F", code="G")
    with pytest.raises(IntegrityError):
        ActivityParameter.create(group="E", database="E", name="F", code="G")


##########
### Groups
##########


@bw2test
def test_group():
    o = Group.create(name="foo")
    assert o.fresh
    o.expire()
    assert not o.fresh
    o = Group.get(name="foo")
    assert not o.fresh
    o.freshen()
    assert o.fresh
    o = Group.get(name="foo")
    assert o.fresh
    with pytest.raises(IntegrityError):
        Group.create(name="foo")
    Group.create(name="bar")


@bw2test
def test_group_purging():
    Database("A").register()
    Database("B").register()
    o = Group.create(name="one", order=["C", "project", "B", "D", "A"])
    expected = ["C", "D"]
    assert o.updated
    assert o.fresh
    assert o.order == expected
    assert Group.get(name="one").order == expected


######################
### Group dependencies
######################


@bw2test
def test_group_dependency():
    d = GroupDependency.create(group="foo", depends="bar")
    assert d.group == "foo"
    assert d.depends == "bar"


@bw2test
def test_group_dependency_save_checks():
    with pytest.raises(ValueError):
        GroupDependency.create(group="project", depends="foo")
    Database("A").register()
    GroupDependency.create(group="A", depends="project")
    with pytest.raises(ValueError):
        GroupDependency.create(group="A", depends="foo")


@bw2test
def test_group_dependency_constraints():
    GroupDependency.create(group="foo", depends="bar")
    with pytest.raises(IntegrityError):
        GroupDependency.create(group="foo", depends="bar")
    with pytest.raises(IntegrityError):
        GroupDependency.create(group="foo", depends="foo")


@bw2test
def test_group_dependency_circular():
    GroupDependency.create(group="foo", depends="bar")
    with pytest.raises(IntegrityError):
        GroupDependency.create(group="bar", depends="foo")


@bw2test
def test_group_dependency_override():
    """ GroupDependency can be overridden by having a parameter with the same
    name within the group.
    """
    ProjectParameter.create(name="foo", amount=2)
    Database("B").register()
    DatabaseParameter.create(database="B", name="bar", amount=1, formula="foo * 5")
    parameters.recalculate()
    assert (
        GroupDependency.select().where(GroupDependency.depends == "project").count()
        == 1
    )
    assert DatabaseParameter.get(name="bar").amount == 10
    DatabaseParameter.create(
        database="B", name="foo", amount=8,
    )
    parameters.recalculate()
    assert (
        GroupDependency.select().where(GroupDependency.depends == "project").count()
        == 0
    )
    assert DatabaseParameter.get(name="bar").amount == 40


######################
### Parameters manager
######################


@bw2test
def test_parameters_new_project_parameters_uniqueness():
    with pytest.raises(AssertionError):
        parameters.new_project_parameters([{"name": "foo"}, {"name": "foo"}])


@bw2test
def test_parameters_new_project_parameters():
    assert not len(parameters)
    ProjectParameter.create(name="foo", amount=17)
    ProjectParameter.create(name="baz", amount=10)
    assert len(parameters) == 2
    assert ProjectParameter.get(name="foo").amount == 17
    data = [
        {"name": "foo", "amount": 4},
        {"name": "bar", "formula": "foo + 3"},
    ]
    parameters.new_project_parameters(data)
    assert len(parameters) == 3
    assert ProjectParameter.get(name="foo").amount == 4
    assert ProjectParameter.get(name="bar").amount == 7
    assert ProjectParameter.get(name="baz").amount == 10
    assert Group.get(name="project").fresh


@bw2test
def test_parameters_new_project_parameters_no_overwrite():
    ProjectParameter.create(name="foo", amount=17)
    data = [
        {"name": "foo", "amount": 4},
        {"name": "bar", "formula": "foo + 3"},
    ]
    with pytest.raises(ValueError):
        parameters.new_project_parameters(data, overwrite=False)


@bw2test
def test_parameters_repr():
    assert repr(parameters) == "Parameters manager with 0 objects"


@bw2test
def test_parameters_recalculate():
    Database("B").register()
    ActivityParameter.create(
        group="A", database="B", code="C", name="D", formula="2 ** 3"
    )
    ActivityParameter.create(
        group="A", database="B", code="E", name="F", formula="foo + bar + D"
    )
    DatabaseParameter.create(
        database="B", name="foo", formula="2 ** 2",
    )
    ProjectParameter.create(
        name="bar", formula="2 * 2 * 2",
    )
    parameters.recalculate()
    assert ProjectParameter.get(name="bar").amount == 8
    assert DatabaseParameter.get(name="foo").amount == 4
    assert ActivityParameter.get(name="F").amount == 20
    assert ActivityParameter.get(name="D").amount == 8


@bw2test
def test_parameters_new_database_parameters():
    with pytest.raises(AssertionError):
        parameters.new_database_parameters([], "another")
    Database("another").register()
    with pytest.raises(AssertionError):
        parameters.new_database_parameters(
            [{"name": "foo"}, {"name": "foo"}], "another"
        )
    DatabaseParameter.create(name="foo", database="another", amount=0)
    DatabaseParameter.create(name="baz", database="another", amount=21)
    assert len(parameters) == 2
    assert DatabaseParameter.get(name="foo").amount == 0
    data = [
        {"name": "foo", "amount": 4},
        {"name": "bar", "formula": "foo + 3"},
    ]
    parameters.new_database_parameters(data, "another")
    assert len(parameters) == 3
    assert DatabaseParameter.get(name="foo").amount == 4
    assert DatabaseParameter.get(name="bar").amount == 7
    assert DatabaseParameter.get(name="baz").amount == 21
    assert Group.get(name="another").fresh


@bw2test
def test_parameters_new_database_parameters_no_overwrite():
    Database("another").register()
    DatabaseParameter.create(name="foo", database="another", amount=0)
    with pytest.raises(ValueError):
        parameters.new_database_parameters(
            [{"name": "foo", "amount": 4}], "another", overwrite=False
        )


@bw2test
def test_parameters_new_activity_parameters_errors():
    with pytest.raises(AssertionError):
        parameters.new_activity_parameters([], "example")
    with pytest.raises(AssertionError):
        parameters.new_activity_parameters(
            [{"database": 1}, {"database": 2}], "example"
        )

    with pytest.raises(AssertionError):
        parameters.new_activity_parameters([{"database": "unknown"}], "example")

    Database("A").register()
    with pytest.raises(AssertionError):
        parameters.new_activity_parameters(
            [{"database": "A", "name": "foo"}, {"database": "A", "name": "foo"}],
            "example",
        )


@bw2test
def test_parameters_new_activity_parameters():
    assert not len(parameters)
    assert not Group.select().count()
    Database("A").register()
    ActivityParameter.create(
        group="another", database="A", name="baz", code="D", amount=49
    )
    ActivityParameter.create(
        group="another", database="A", name="foo", code="E", amount=101
    )
    assert len(parameters) == 2
    assert ActivityParameter.get(name="foo").amount == 101
    assert ActivityParameter.get(name="baz").amount == 49
    data = [
        {"database": "A", "code": "B", "name": "foo", "amount": 4},
        {
            "database": "A",
            "code": "C",
            "name": "bar",
            "formula": "foo + 3",
            "uncertainty type": 0,
        },
    ]
    parameters.new_activity_parameters(data, "another")
    assert len(parameters) == 3
    assert ActivityParameter.get(name="foo").amount == 4
    assert ActivityParameter.get(name="foo").code == "B"
    assert ActivityParameter.get(name="baz").amount == 49
    a = ActivityParameter.get(code="C")
    assert a.database == "A"
    assert a.name == "bar"
    assert a.formula == "foo + 3"
    assert a.data == {"uncertainty type": 0}
    assert a.amount == 7
    assert ActivityParameter.get(name="foo").amount == 4
    assert Group.get(name="another").fresh


@bw2test
def test_parameters_new_activity_parameters_no_overlap():
    Database("A").register()
    ActivityParameter.create(
        group="another", database="A", name="foo", code="D", amount=49
    )
    data = [
        {"database": "A", "code": "B", "name": "foo", "amount": 4},
        {
            "database": "A",
            "code": "C",
            "name": "bar",
            "formula": "foo + 3",
            "uncertainty type": 0,
        },
    ]
    with pytest.raises(ValueError):
        parameters.new_activity_parameters(data, "another", overwrite=False)


@bw2test
def test_parameters_rename_project_parameter():
    """ Project parameters can be renamed. """
    param = ProjectParameter.create(name="foo", amount=7,)
    assert ProjectParameter.select().where(ProjectParameter.name == "foo").count() == 1
    parameters.rename_project_parameter(param, "foobar")
    with pytest.raises(ProjectParameter.DoesNotExist):
        ProjectParameter.get(name="foo")
    assert (
        ProjectParameter.select().where(ProjectParameter.name == "foobar").count() == 1
    )


@bw2test
def test_parameters_rename_project_parameter_incorrect_type():
    Database("B").register()
    param = DatabaseParameter.create(database="B", name="foo", amount=5,)
    with pytest.raises(TypeError):
        parameters.rename_project_parameter(param, "bar")


@bw2test
def test_parameters_rename_project_parameter_dependencies():
    """ Updating downstream parameters will update all relevant formulas
    to use the new name for the parameter.
    """
    param = ProjectParameter.create(name="foo", amount=7,)
    ProjectParameter.create(name="bar", amount=1, formula="foo * 2")
    assert ProjectParameter.is_dependency_within_group("foo")
    parameters.rename_project_parameter(param, "baz", update_dependencies=True)
    assert ProjectParameter.get(name="bar").formula == "baz * 2"


@bw2test
def test_parameters_rename_project_parameter_dependencies_fail():
    """ An exception is raised if rename is attempted without updating
    downstream if other parameters depend on that parameter.
    """
    param = ProjectParameter.create(name="foo", amount=7,)
    ProjectParameter.create(name="bar", amount=1, formula="foo * 2")
    with pytest.raises(ValueError):
        parameters.rename_project_parameter(param, "baz")


def test_parameters_rename_project_parameter_dependencies_full(chain):
    """ Updating downstream parameters will update all relevant formulas
    to use the new name for the parameter.

    Parameter amounts do no change as only the name is altered.
    """
    ProjectParameter.create(name="double_bar", amount=12, formula="bar * 2")
    DatabaseParameter.create(database="B", name="bing", amount=2, formula="bar ** 5")
    parameters.recalculate()
    assert ProjectParameter.is_dependency_within_group("bar")
    assert DatabaseParameter.is_dependent_on("bar")
    assert ActivityParameter.is_dependent_on("bar", "project")
    assert ProjectParameter.get(name="double_bar").amount == 16
    assert DatabaseParameter.get(name="bing", database="B").amount == 32768
    assert ActivityParameter.get(name="F", group="A").amount == 20

    param = ProjectParameter.get(name="bar")
    parameters.rename_project_parameter(param, "new_bar", update_dependencies=True)

    assert ProjectParameter.get(name="double_bar").formula == "new_bar * 2"
    assert DatabaseParameter.get(name="bing", database="B").formula == "new_bar ** 5"
    assert ActivityParameter.get(name="F", group="A").formula == "foo + new_bar + D"
    assert ProjectParameter.get(name="double_bar").amount == 16
    assert DatabaseParameter.get(name="bing", database="B").amount == 32768
    assert ActivityParameter.get(name="F", group="A").amount == 20


@bw2test
def test_parameters_rename_database_parameter():
    Database("B").register()
    param = DatabaseParameter.create(database="B", name="foo", amount=5,)
    assert (
        DatabaseParameter.select().where(DatabaseParameter.name == "foo").count() == 1
    )
    parameters.rename_database_parameter(param, "bar")
    with pytest.raises(DatabaseParameter.DoesNotExist):
        DatabaseParameter.get(name="foo")
    assert (
        DatabaseParameter.select().where(DatabaseParameter.name == "bar").count() == 1
    )


def test_parameters_rename_database_parameter_dependencies(chain):
    DatabaseParameter.create(database="B", name="baz", amount=1, formula="foo + 2")
    parameters.recalculate()
    param = DatabaseParameter.get(name="foo")
    parameters.rename_database_parameter(param, "foobar", True)
    assert DatabaseParameter.get(name="baz").formula == "foobar + 2"
    assert ActivityParameter.get(name="F", group="A").formula == "foobar + bar + D"


def test_parameters_rename_activity_parameter(chain):
    parameters.recalculate()
    param = ActivityParameter.get(name="J", group="G")
    parameters.rename_activity_parameter(param, "John")
    with pytest.raises(ActivityParameter.DoesNotExist):
        ActivityParameter.get(name="J", group="G")
    assert (
        ActivityParameter.select()
        .where(ActivityParameter.name == "John", ActivityParameter.group == "G")
        .count()
        == 1
    )


def test_parameters_rename_activity_parameter_dependencies(chain):
    parameters.recalculate()
    param = ActivityParameter.get(name="D", group="A")
    parameters.rename_activity_parameter(param, "Dirk", True)
    assert ActivityParameter.get(name="F", group="A").formula == "foo + bar + Dirk"
    assert ActivityParameter.get(name="J", group="G").formula == "F + Dirk * 2"


@bw2test
def test_parameters_rename_activity_parameter_group_exchange():
    """ Rename 'D' from group 'A' updates ParameterizedExchange and
    underlying exchange.
    """
    db = Database("B")
    db.register()
    ActivityParameter.create(
        group="A", database="B", code="C", name="D", formula="2 ** 3", amount=1,
    )
    a = db.new_activity(code="newcode", name="new activity")
    a.save()
    a.new_exchange(amount=1, input=a, type="production", formula="D + 2").save()
    parameters.add_exchanges_to_group("A", a)
    ActivityParameter.recalculate_exchanges("A")

    param = ActivityParameter.get(name="D", group="A")
    parameters.rename_activity_parameter(param, "Correct", True)
    assert ParameterizedExchange.get(group="A").formula == "Correct + 2"
    exc = next(iter(a.production()))
    assert exc.amount == 10
    assert exc.get("formula") == "Correct + 2"


@bw2test
def test_parameters_rename_activity_parameter_order_exchange():
    """ Rename 'D' from group 'A' updates ParameterizedExchange and
    underlying exchange in group 'G'
    """
    db = Database("K")
    db.register()
    ActivityParameter.create(
        group="A", database="K", code="C", name="D", formula="2 ** 3", amount=1,
    )
    a = db.new_activity(code="newcode", name="new activity")
    a.save()
    a.new_exchange(amount=1, input=a, type="production", formula="D + 2").save()
    Group.create(name="G", order=["A"], fresh=False)
    parameters.add_exchanges_to_group("G", a)
    ActivityParameter.recalculate_exchanges("G")

    param = ActivityParameter.get(name="D", group="A")
    parameters.rename_activity_parameter(param, "Correct", update_dependencies=True)

    assert ParameterizedExchange.get(group="G").formula == "Correct + 2"
    exc = next(iter(a.production()))
    assert exc.amount == 10
    assert exc.get("formula") == "Correct + 2"


@bw2test
def test_parameters_add_to_group_empty():
    db = Database("example")
    db.register()
    assert not len(parameters)
    assert not len(db)
    assert not Group.select().count()
    a = db.new_activity(code="A", name="An activity",)
    a.save()
    assert parameters.add_to_group("my group", a) is None
    assert Group.get(name="my group")
    assert not len(parameters)


@bw2test
def test_parameters_add_to_group():
    db = Database("example")
    db.register()
    assert not len(parameters)
    assert not len(db)
    assert not Group.select().count()

    ActivityParameter.create(
        group="my group", database="example", name="bye bye", code="A", amount=1,
    )

    a = db.new_activity(
        code="A",
        name="An activity",
        parameters=[
            {"amount": 4, "name": "one", "foo": "bar"},
            {"amount": 42, "name": "two", "formula": "this + that"},
        ],
    )
    a.save()
    assert "parameters" in get_activity(("example", "A"))

    assert parameters.add_to_group("my group", a) == 2
    assert Group.get(name="my group")
    assert (
        not ActivityParameter.select()
        .where(ActivityParameter.name == "bye bye")
        .count()
    )
    expected = (
        ("one", 4, None, {"foo": "bar"}),
        ("two", 42, "this + that", {}),
    )
    for ap in ActivityParameter.select():
        assert (ap.name, ap.amount, ap.formula, ap.data) in expected
    assert "parameters" not in get_activity(("example", "A"))


@bw2test
def test_parameters_remove_from_group():
    db = Database("example")
    db.register()
    a = db.new_activity(code="A", name="An activity")
    a.save()
    b = db.new_activity(code="B", name="Another activity")
    b.save()
    a.new_exchange(amount=0, input=b, type="technosphere", formula="bar + 4").save()
    activity_data = [
        {
            "name": "reference_me",
            "formula": "sqrt(25)",
            "database": "example",
            "code": "B",
        },
        {
            "name": "bar",
            "formula": "reference_me + 2",
            "database": "example",
            "code": "A",
        },
    ]
    parameters.new_activity_parameters(activity_data, "my group")
    parameters.add_exchanges_to_group("my group", a)
    assert not get_activity(("example", "A")).get("parameters")
    assert ActivityParameter.select().count() == 2
    assert ParameterizedExchange.select().count() == 1

    parameters.remove_from_group("my group", a)
    assert ActivityParameter.select().count() == 1
    assert not ParameterizedExchange.select().count()
    assert get_activity(("example", "A"))["parameters"]


@bw2test
def test_parameters_save_restore_exchange_amount():
    """ The original amount of the exchange is restored when it is no
    longer parameterized.
    """
    db = Database("example")
    db.register()
    a = db.new_activity(code="A", name="An activity")
    a.save()
    b = db.new_activity(code="B", name="Another activity")
    b.save()
    a.new_exchange(amount=5, input=b, type="technosphere", formula="bing + 5").save()

    activity_data = [
        {"name": "bing", "amount": "7", "database": "example", "code": "A",}
    ]
    parameters.new_activity_parameters(activity_data, "calculate")
    parameters.add_exchanges_to_group("calculate", a)
    # The original amount and current amount is 5
    for exc in a.exchanges():
        assert exc["amount"] == 5
        assert "original_amount" in exc and exc["original_amount"] == 5

    ActivityParameter.recalculate_exchanges("calculate")
    # Parameterization has caused the amount to change.
    for exc in a.exchanges():
        assert exc["amount"] == 12
        assert "original_amount" in exc

    # Remove parameterization from the activity, restoring the original amount
    parameters.remove_from_group("calculate", a)
    for exc in a.exchanges():
        assert exc["amount"] == 5
        assert "original_amount" not in exc


@bw2test
def test_parameters_save_keep_changed_exchange_amount():
    db = Database("example")
    db.register()
    a = db.new_activity(code="A", name="An activity")
    a.save()
    b = db.new_activity(code="B", name="Another activity")
    b.save()
    a.new_exchange(amount=5, input=b, type="technosphere", formula="bing + 5").save()

    activity_data = [
        {"name": "bing", "amount": "7", "database": "example", "code": "A",}
    ]
    parameters.new_activity_parameters(activity_data, "calculate")
    parameters.add_exchanges_to_group("calculate", a)
    # The original amount and current amount is 5
    for exc in a.exchanges():
        assert exc["amount"] == 5
        assert "original_amount" in exc and exc["original_amount"] == 5

    ActivityParameter.recalculate_exchanges("calculate")
    # Parameterization has caused the amount to change.
    for exc in a.exchanges():
        assert exc["amount"] == 12
        assert "original_amount" in exc

    # Remove parameterization from the activity, keeping the changed amount
    parameters.remove_from_group("calculate", a, restore_amounts=False)
    for exc in a.exchanges():
        assert exc["amount"] == 12
        assert "original_amount" in exc



from . import methods, geomapping, config
from .ia_data_store import ImpactAssessmentDataStore
from .utils import as_uncertainty_dict
from .validate import ia_validator
from .backends.schema import get_id


class Method(ImpactAssessmentDataStore):
    """A manager for an impact assessment method. This class can register or deregister methods, write intermediate data, process data to parameter arrays, validate, and copy methods.

    The Method class never holds intermediate data, but it can load or write intermediate data. The only attribute is *name*, which is the name of the method being managed.

    Instantiation does not load any data. If this method is not yet registered in the metadata store, a warning is written to ``stdout``.

    Methods are hierarchally structured, and this structure is preserved in the method name. It is a tuple of strings, like ``('ecological scarcity 2006', 'total', 'natural resources')``.

    The data schema for IA methods is:

    .. code-block:: python

            Schema([Any(
                [valid_tuple, maybe_uncertainty],         # site-generic
                [valid_tuple, maybe_uncertainty, object]  # regionalized
            )])

    where:
        * *valid_tuple* (tuple): A dataset identifier, like ``("biosphere", "CO2")``.
        * *maybe_uncertainty* (uncertainty dict or number): Either a number or an uncertainty dictionary.
        * *object* (object, optional) is a location identifier, used only for regionalized LCIA.

    Args:
        * *name* (tuple): Name of impact assessment method to manage.

    """

    _metadata = methods
    validator = ia_validator
    matrix = "characterization_matrix"

    def add_geomappings(self, data):
        geomapping.add({x[2] for x in data if len(x) == 3})

    def process_row(self, row):
        """Given ``(flow, amount, maybe location)``, return a dictionary for array insertion."""
        return {
            **as_uncertainty_dict(row[1]),
            "row": get_id(row[0]),
            "col": (
                geomapping[row[2]]
                if len(row) >= 3
                else geomapping[config.global_location]
            ),
        }

    def write(self, data, process=True):
        """Serialize intermediate data to disk.

        Sets the metadata key ``num_cfs`` automatically."""
        if self.name not in self._metadata:
            self.register()
        self.metadata["num_cfs"] = len(data)
        self._metadata.flush()
        super(Method, self).write(data)

    def process(self, **extra_metadata):
        extra_metadata["global_index"] = geomapping[config.global_location]
        super().process(**extra_metadata)



from . import config, projects
from .utils import random_string, create_in_memory_zipfile_from_directory
from logging.handlers import RotatingFileHandler
import codecs
import datetime
import logging
import requests
import uuid

try:
    import anyjson
except ImportError:
    anyjson = None


class FakeLog(object):
    """Like a log object, but does nothing"""

    def fake_function(cls, *args, **kwargs):
        return

    def __getattr__(self, attr):
        return self.fake_function


def get_logger(name, level=logging.INFO):
    filename = "{}-{}.log".format(
        name, datetime.datetime.now().strftime("%d-%B-%Y-%I-%M%p"),
    )
    handler = RotatingFileHandler(
        projects.logs_dir / filename, maxBytes=1e6, encoding="utf-8", backupCount=10,
    )
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(lineno)d %(message)s")
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def get_io_logger(name):
    """Build a logger that records only relevent data for display later as HTML."""
    filepath = projects.logs_dir / "{}.{}.log".format(name, random_string(6))
    handler = logging.StreamHandler(codecs.open(filepath, "w", "utf-8"))
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(handler)
    return logger, filepath


def get_verbose_logger(name, level=logging.WARNING):
    filename = "{}-{}.log".format(
        name, datetime.datetime.now().strftime("%d-%B-%Y-%I-%M%p"),
    )
    handler = RotatingFileHandler(
        projects.logs_dir / filename, maxBytes=50000, encoding="utf-8", backupCount=5,
    )
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(level)
    handler.setFormatter(
        logging.Formatter(
            """
Message type:       %(levelname)s
Location:           %(pathname)s:%(lineno)d
Module:             %(module)s
Function:           %(funcName)s
Time:               %(asctime)s
Message:
%(message)s

"""
        )
    )
    logger.addHandler(handler)
    return logger


def upload_logs_to_server(metadata={}):
    # Hardcoded for now
    url = "http://reports.brightwaylca.org/logs"
    zip_fo = create_in_memory_zipfile_from_directory(projects.logs_dir)
    files = {"file": (uuid.uuid4().hex + ".zip", zip_fo.read())}
    metadata["json"] = "native" if anyjson is None else anyjson.implementation.name
    metadata["windows"] = config._windows
    return requests.post(url, data=metadata, files=files)


def close_log(log):
    """Detach log handlers; flush to disk"""
    handlers = log.handlers[:]
    for handler in handlers:
        handler.close()
        log.removeHandler(handler)



from setuptools import setup
import os

packages = []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('bw2data'):
    # Ignore dirnames that start with '.'
    if '__init__.py' in filenames:
        pkg = dirpath.replace(os.path.sep, '.')
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, '.')
        packages.append(pkg)

v_temp = {}
with open("bw2data/version.py") as fp:
    exec(fp.read(), v_temp)
version = ".".join((str(x) for x in v_temp["version"]))

setup(
    name='bw2data',
    version=version,
    packages=packages,
    python_requires='>=3.5',
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license="3-clause BSD",
    install_requires=[
        "appdirs",
        "bw2parameters",
        "bw_processing",
        "docopt",
        "fasteners",
        "lxml",
        "numpy",
        "peewee>=3.9.4",
        "psutil",
        "pyprind",
        "requests>=1.1.0",
        "scipy",
        "stats_arrays",
        "voluptuous",
        "whoosh",
        "wrapt",
    ],
    url="https://github.com/brightway-lca/brightway2-data",
    long_description=open('README.rst').read(),
    description=('Tools for the management of inventory databases '
                 'and impact assessment methods. Part of the Brightway2 LCA Framework'),
    entry_points = {
        'console_scripts': [
            'bw2-uptodate = bw2data.bin.bw2_uptodate:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],)
